# -*- coding: utf-8 -*-
# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = "BackupCleaner"

import os
import re

from iconcommons.logger import Logger
from . import get_backup_filename
from ..icon_constant import BACKUP_LOG_TAG, BACKUP_FILES

_TAG = BACKUP_LOG_TAG


class BackupCleaner(object):
    """Remove old backup files to rollback

    """

    def __init__(self, backup_root_path: str, backup_files: int):
        """

        :param backup_root_path: the directory where backup files are placed
        :param backup_files: the maximum backup files to keep
        """
        self._backup_root_path = backup_root_path
        self._backup_files = backup_files if backup_files > 0 else BACKUP_FILES
        self._regex_object = re.compile("^[\d]{10}.bak$")

    def run_on_init(self, current_block_height: int) -> int:
        """Clean up all stale backup files on iconservice startup

        :param current_block_height:
        :return:
        """
        Logger.debug(tag=_TAG, msg=f"run_on_init() start")

        ret = 0
        start_block_height = max(0, current_block_height - self._backup_files)

        with os.scandir(self._backup_root_path) as it:
            for entry in it:
                # backup filename: ex) 0000012345.bak
                if entry.is_file() and self._is_backup_filename_valid(entry.name):
                    block_height: int = self._get_block_height_from_filename(entry.name)
                    if block_height < 0:
                        continue

                    # Do nothing for the latest backup files
                    if start_block_height <= block_height < current_block_height:
                        continue

                    # Remove stale backup files
                    if self._remove_file(entry.path):
                        ret += 1

        Logger.debug(tag=_TAG, msg=f"run_on_init() end: ret={ret}")
        return ret

    @staticmethod
    def _get_block_height_from_filename(filename: str) -> int:
        try:
            return int(filename[:-4])
        except:
            pass

        return -1

    def _is_backup_filename_valid(self, filename: str) -> bool:
        return len(filename) == 14 and bool(self._regex_object.match(filename))

    def run_on_commit(self, current_block_height: int) -> int:
        """Remove the oldest backup file on commit

        :param: current_block_height
        :param: func: function to remove a file with path
        :return: the number of removed files
        """
        Logger.debug(tag=_TAG,
                     msg=f"run() start: current_block_height={current_block_height} "
                         f"backup_files={self._backup_files}")

        # Remove the oldest backup file only
        start_block_height = current_block_height - self._backup_files - 1
        ret = self.run(start_block_height, end_block_height=start_block_height)

        Logger.debug(tag=_TAG, msg="run() end")

        return ret

    def run(self, start_block_height: int, end_block_height: int) -> int:
        """Remove block backup files ranging from start_block_height to end_block_height inclusive

        :param start_block_height:
        :param end_block_height:
        :return: The number of removed files
        """
        Logger.debug(tag=_TAG, msg=f"run() start: start={start_block_height} end={end_block_height}")

        # Parameters sanity check
        if start_block_height < 0 or start_block_height > end_block_height:
            Logger.warning(tag=_TAG, msg=f"Invalid range: start={start_block_height} end={end_block_height}")
            return -1

        ret = 0

        # Remove block backup files ranging from start_block_height to end_block_height inclusive
        for block_height in range(start_block_height, end_block_height + 1):
            filename: str = get_backup_filename(block_height)
            path: str = os.path.join(self._backup_root_path, filename)

            # Remove a file ignoring any exceptions
            if self._remove_file(path):
                ret += 1

        Logger.info(tag=_TAG,
                    msg=f"Clean up old backup files: "
                        f"start={start_block_height} end={end_block_height} count={ret}")
        Logger.debug(tag=_TAG, msg=f"run() end: ret={ret}")

        return ret

    @staticmethod
    def _remove_file(path: str) -> bool:
        try:
            os.remove(path)
            return True
        except FileNotFoundError:
            pass
        except BaseException as e:
            Logger.warning(tag=_TAG, msg=str(e))

        return False



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import json
import os
import sys
from typing import TYPE_CHECKING

import iconservice.iconscore.utils as utils
from iconservice.base.address import Address
from iconservice.base.exception import IllegalFormatException
from iconservice.icon_constant import PACKAGE_JSON_FILE

if TYPE_CHECKING:
    from iconservice.base.address import Address


class IconScoreClassLoader:
    """IconScoreBase subclass Loader

    """
    @classmethod
    def init(cls, score_root_path: str):
        if score_root_path not in sys.path:
            sys.path.append(score_root_path)

    @classmethod
    def close(cls, score_root_path: str):
        sys.path.remove(score_root_path)

    @classmethod
    def _load_package_json(cls, score_deploy_path: str) -> dict:
        """Loads package.json in SCORE

        :param score_deploy_path:
        :return:
        """
        pkg_json_path = os.path.join(score_deploy_path, PACKAGE_JSON_FILE)
        with open(pkg_json_path, 'r') as f:
            return json.load(f)

    @classmethod
    def _get_package_info(cls, package_json: dict) -> tuple:
        """Returns main_module and main_score

        :param package_json: dict returned by _load_package_json()
        :return: tuple containing main_module and main_score
        """
        main_module: str = package_json.get('main_module')
        if not isinstance(main_module, str):
            # "main_file" field will be deprecated soon.
            # Use "main_module" instead
            main_module: str = package_json['main_file']

        # Relative package name is not allowed
        if main_module.startswith('.'):
            raise IllegalFormatException('Invalid main_module')

        main_score: str = package_json['main_score']

        return main_module, main_score

    @classmethod
    def run(cls, score_address: 'Address', tx_hash: bytes, score_root_path: str) -> type:
        """Load a IconScoreBase subclass and return it

        :param score_address:
        :param tx_hash:
        :param score_root_path:
        :return: subclass derived from IconScoreBase
        """
        score_deploy_path: str = utils.get_score_deploy_path(score_root_path, score_address, tx_hash)
        package_name: str = utils.get_package_name_by_address_and_tx_hash(score_address, tx_hash)

        package_json: dict = IconScoreClassLoader._load_package_json(score_deploy_path)
        main_module, main_score = IconScoreClassLoader._get_package_info(package_json)

        # In order for the new module to be noticed by the import system
        importlib.invalidate_caches()
        module = importlib.import_module(f".{main_module}", package_name)

        return getattr(module, main_score)



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Any

from iconcommons.logger import Logger

from .icon_score_step import get_input_data_size
from ..base.address import Address, SYSTEM_SCORE_ADDRESS, generate_score_address
from ..base.exception import InvalidRequestException, InvalidParamsException, OutOfBalanceException
from ..icon_constant import FIXED_FEE, MAX_DATA_SIZE, DEFAULT_BYTE_SIZE, DATA_BYTE_ORDER, Revision, DeployState
from ..utils import is_lowercase_hex_string

if TYPE_CHECKING:
    from ..deploy.storage import IconScoreDeployInfo
    from .icon_score_context import IconScoreContext


TAG = "PV"


class IconPreValidator:
    """Validate only icx_sendTransaction request before putting it into tx pool

    It does not validate query requests like icx_getBalance, icx_call and so on
    """

    def __init__(self) -> None:
        """Constructor
        """
        pass

    def execute(self, context: 'IconScoreContext', params: dict, step_price: int, minimum_step: int):
        """Validate a transaction on icx_sendTransaction
        If failed to validate a tx, raise an exception

        Assume that values in params have already been converted
        to original format (string -> int, string -> Address, etc)

        :param context:
        :param params: params of icx_sendTransaction JSON-RPC request
        :param step_price:
        :param minimum_step: minimum step
        """

        self._check_input_data(params)

        value: int = params.get('value', 0)
        if value < 0:
            raise InvalidParamsException("value < 0")
        try:
            value.to_bytes(DEFAULT_BYTE_SIZE, DATA_BYTE_ORDER)
        except OverflowError:
            raise InvalidParamsException("exceed ICX amount you can send at one time")

        version: int = params.get('version', 2)
        if version < 3:
            self._validate_transaction_v2(context, params)
        else:
            self._validate_transaction_v3(context, params, step_price, minimum_step)

    def execute_to_check_out_of_balance(self, context: 'IconScoreContext', params: dict, step_price: int):
        version: int = params.get('version', 2)

        if version < 3:
            self._check_from_can_charge_fee_v2(context, params)
        else:
            self._check_from_can_charge_fee_v3(context, params, step_price)

    @staticmethod
    def _check_input_data(params: dict):
        """
        Validates input data. It checks the input data type and the input data size.

        :param params: params of icx_sendTransaction JSON-RPC request
        :return:
        """

        input_data = params.get('data', None)
        if 'message' == params.get('dataType', None):
            IconPreValidator._check_message_data(input_data)
        else:
            IconPreValidator._check_input_data_type(input_data)

        IconPreValidator._check_input_data_size(input_data)

    @staticmethod
    def _check_message_data(data: Any):
        """
        Check if the message data is a lowercase hex string

        :param data: input data of message type
        """
        if isinstance(data, str) \
                and data.startswith('0x') \
                and is_lowercase_hex_string(data[2:]) \
                and len(data) % 2 == 0:
            return

        raise InvalidRequestException('Invalid message data')

    @staticmethod
    def _check_input_data_type(data: Any):
        """
        Validates transaction data types whether the leaf fields are str or None
        """
        if isinstance(data, dict):
            for v in data.values():
                IconPreValidator._check_input_data_type(v)
        elif isinstance(data, list):
            for v in data:
                IconPreValidator._check_input_data_type(v)
        elif data is not None and not isinstance(data, str):
            # The leaf value should be None or str.
            raise InvalidRequestException('Invalid data type')

    @staticmethod
    def _check_input_data_size(input_data: Any):
        """
        Validates transaction data whether total bytes is less than MAX_DATA_SIZE
        If the property is a key-value object, counts key and value.

        Assume that values in params have already been converted
        to original format (string -> int, string -> Address, etc)
        But the field of 'data' has not been converted (TypeConvert marks it as LATER)

        :param input_data: data field of icx_sendTransaction JSON-RPC request
        """

        if input_data is not None:
            size = get_input_data_size(Revision.LATEST.value, input_data)

            if size > MAX_DATA_SIZE:
                raise InvalidRequestException('Invalid message length')

    def _check_from_can_charge_fee_v2(self, context: 'IconScoreContext', params: dict):
        fee: int = params['fee']
        if fee != FIXED_FEE:
            raise InvalidRequestException(f'Invalid fee: {fee}')

        from_: 'Address' = params['from']
        value: int = params.get('value', 0)

        self._check_balance(context, from_, value, fee)

    def _validate_transaction_v2(self, context: 'IconScoreContext', params: dict):
        """Validate transfer transaction based on protocol v2

        :param params:
        :return:
        """
        # Check out of balance
        self._check_from_can_charge_fee_v2(context, params)

        # Check 'to' is not a SCORE address
        to: 'Address' = params['to']
        if to.is_contract:
            raise InvalidRequestException(
                'Not allowed to transfer coin to SCORE on protocol v2')

    def _validate_transaction_v3(self, context: 'IconScoreContext', params: dict, step_price: int, minimum_step: int):
        """Validate transfer transaction based on protocol v3

        :param params:
        :return:
        """
        self._check_minimum_step(params, minimum_step)
        self._check_from_can_charge_fee_v3(context, params, step_price)

        # Check if "to" address is valid
        to: 'Address' = params['to']

        if self._is_inactive_score(context, to):
            raise InvalidRequestException(f'{to} is inactive SCORE')

        # Check data_type-specific elements
        data_type = params.get('dataType', None)
        if data_type == 'call':
            self._validate_call_transaction(context, params)
        elif data_type == 'deploy':
            self._validate_deploy_transaction(context, params)
        elif data_type == 'deposit':
            self._validate_deposit_transaction(context, params)

    @staticmethod
    def _check_minimum_step(params: dict, minimum_step: int):
        step_limit = params.get('stepLimit', 0)
        if step_limit < minimum_step:
            raise InvalidRequestException('Step limit too low')

    def _check_from_can_charge_fee_v3(self, context: 'IconScoreContext', params: dict, step_price: int):
        from_: 'Address' = params['from']
        to: 'Address' = params['to']
        value: int = params.get('value', 0)

        step_limit = params.get('stepLimit', 0)
        fee = step_limit * step_price

        self._check_balance(context, from_, value, fee)

        data_type: str = params.get('dataType')
        if to.is_contract and data_type in (None, 'call', 'message'):
            # Check if the SCORE can be called when fee-sharing ON.
            # If data_type is None or message and the recipient is SCORE,
            # it works like `call`.(calling fallback)
            context.engine.fee.check_score_available(context, to, context.block.height)

    def _validate_call_transaction(self, context: 'IconScoreContext', params: dict):
        """Validate call transaction
        It is not icx_call

        :param params:
        :return:
        """
        to: 'Address' = params['to']

        if self._is_inactive_score(context, to):
            raise InvalidRequestException(f'{to} is inactive SCORE')

        data = params.get('data', None)
        if not isinstance(data, dict):
            raise InvalidRequestException('Data not found')

        if 'method' not in data:
            raise InvalidRequestException('Method not found')

    def _validate_deploy_transaction(self, context: 'IconScoreContext', params: dict):
        to: 'Address' = params['to']

        value: int = params.get('value', 0)
        if value != 0:
            raise InvalidParamsException('value must be 0 in a deploy transaction')

        if self._is_inactive_score(context, to):
            raise InvalidRequestException(f'{to} is an inactive SCORE')

        data = params.get('data', None)
        if not isinstance(data, dict):
            raise InvalidRequestException('Data not found')

        if 'contentType' not in data:
            raise InvalidRequestException('ContentType not found')

        if 'content' not in data:
            raise InvalidRequestException('Content not found')

        self._validate_new_score_address_on_deploy_transaction(context, params)

    def _validate_deposit_transaction(self, context: 'IconScoreContext', params: dict):
        """Validate deposit transaction

        :param params:
        :return:
        """
        to: 'Address' = params['to']

        if self._is_inactive_score(context, to):
            raise InvalidRequestException(f'{to} is inactive SCORE')

        data = params.get('data', None)
        if not isinstance(data, dict):
            raise InvalidRequestException('Data not found')

        if 'action' not in data:
            raise InvalidRequestException('Action not found')

    @classmethod
    def _validate_new_score_address_on_deploy_transaction(cls, context: 'IconScoreContext', params: dict):
        """Check if a newly generated score address is available
        Assume that data_type is 'deploy'

        :param params:
        :return:
        """

        to: 'Address' = params['to']
        if to != SYSTEM_SCORE_ADDRESS:
            return

        try:
            data: dict = params['data']
            content_type: str = data['contentType']

            if content_type == 'application/zip':
                from_: 'Address' = params['from']
                timestamp: int = params['timestamp']
                nonce: int = params.get('nonce')

                score_address: 'Address' = generate_score_address(from_, timestamp, nonce)

                deploy_info = context.storage.deploy.get_deploy_info(context, score_address)
                if deploy_info is not None:
                    raise InvalidRequestException(f'SCORE address already in use: {score_address}')
            elif content_type == 'application/tbears':
                pass
            else:
                raise InvalidRequestException(f'Invalid contentType: {content_type}')

        except KeyError as ke:
            raise InvalidParamsException(f'Invalid params: {ke}')
        except BaseException as e:
            raise e

    @classmethod
    def _check_balance(cls, context: 'IconScoreContext', from_: 'Address', value: int, fee: int):
        balance = context.engine.icx.get_balance(context, from_)

        if balance < value + fee:
            msg = f"Out of balance: from={from_} balance={balance} value={value} fee={fee}"
            if balance == 0:
                Logger.info(tag=TAG, msg=f"{msg} {context.block}")

            raise OutOfBalanceException(msg)

    def _is_inactive_score(self, context: 'IconScoreContext', address: 'Address') -> bool:
        is_contract = address.is_contract
        is_zero_score_address = address == SYSTEM_SCORE_ADDRESS
        is_score_active = self._is_score_active(context, address)
        _is_inactive_score = is_contract and not is_zero_score_address and not is_score_active
        return _is_inactive_score

    @classmethod
    def _is_score_active(cls, context: 'IconScoreContext', address: 'Address') -> bool:
        deploy_info: 'IconScoreDeployInfo' = context.storage.deploy.get_deploy_info(context, address)

        if deploy_info is None:
            return False

        return deploy_info.deploy_state == DeployState.ACTIVE



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Optional, Any

from .icon_score_constant import STR_FALLBACK, ATTR_SCORE_CALL
from .icon_score_context_util import IconScoreContextUtil
from .icon_score_event_log import EventLogEmitter
from .icon_score_step import StepType
from .icon_score_trace import Trace, TraceType
from .typing.element import (
    get_score_element_metadata,
    ScoreElementMetadata,
)
from .typing.verification import verify_internal_call_arguments
from ..base.address import Address, SYSTEM_SCORE_ADDRESS, GOVERNANCE_SCORE_ADDRESS
from ..base.exception import StackOverflowException, ScoreNotFoundException
from ..base.message import Message
from ..icon_constant import ICX_TRANSFER_EVENT_LOG, MAX_CALL_STACK_SIZE, IconScoreContextType, Revision

if TYPE_CHECKING:
    from .icon_score_context import IconScoreContext


class InternalCall(object):

    @staticmethod
    def icx_get_balance(context: 'IconScoreContext', address: 'Address') -> int:
        return context.engine.icx.get_balance(context, address)

    @staticmethod
    def other_external_call(context: 'IconScoreContext',
                            addr_from: 'Address',
                            addr_to: 'Address',
                            amount: int,
                            func_name: Optional[str],
                            arg_params: Optional[tuple] = None,
                            kw_params: Optional[dict] = None) -> Any:
        if func_name is None:
            func_name = STR_FALLBACK

        return InternalCall._call(context, addr_from, addr_to, amount, func_name, arg_params, kw_params)

    @staticmethod
    def _call(context: 'IconScoreContext',
              addr_from: 'Address',
              addr_to: 'Address',
              amount: int,
              func_name: str,
              arg_params: Optional[tuple],
              kw_params: Optional[dict]) -> Any:

        InternalCall.enter_call(context)

        try:
            InternalCall._make_trace(context, addr_from, addr_to, amount, func_name, arg_params, kw_params)

            context.step_counter.apply_step(StepType.CONTRACT_CALL, 1)

            context.engine.icx.transfer(context, addr_from, addr_to, amount)

            if amount > 0:
                InternalCall.emit_event_log_for_icx_transfer(context, addr_from, addr_to, amount)

            if addr_to.is_contract:
                # System SCORE inter-call enabled after Revision.SYSTEM_SCORE_ENABLED
                # Exception
                #   - Governance SCORE can call system SCORE inter-call
                if addr_to == SYSTEM_SCORE_ADDRESS and context.revision < Revision.SYSTEM_SCORE_ENABLED.value:
                    if addr_from != GOVERNANCE_SCORE_ADDRESS:
                        raise ScoreNotFoundException(f"{SYSTEM_SCORE_ADDRESS} is not found")

                return InternalCall._other_score_call(
                    context, addr_from, addr_to, amount, func_name, arg_params, kw_params)

            return None
        except BaseException as e:
            InternalCall.revert_call(context)
            raise e
        finally:
            InternalCall.leave_call(context)

    @staticmethod
    def _make_trace(context: 'IconScoreContext',
                    _from: 'Address',
                    _to: 'Address',
                    amount: int,
                    func_name: str,
                    arg_params: Optional[tuple],
                    kw_params: Optional[dict]) -> None:
        if arg_params is None:
            arg_data1 = []
        else:
            arg_data1 = [arg for arg in arg_params]

        if kw_params is None:
            arg_data2 = []
        else:
            arg_data2 = [arg for arg in kw_params.values()]

        arg_data = arg_data1 + arg_data2
        trace = Trace(_from, TraceType.CALL, [_to, func_name, arg_data, amount])
        context.traces.append(trace)

    @staticmethod
    def _other_score_call(context: 'IconScoreContext',
                          addr_from: Address,
                          addr_to: 'Address',
                          amount: int,
                          func_name: str,
                          arg_params: Optional[tuple],
                          kw_params: Optional[dict]) -> Any:
        """Call the functions provided by other icon scores.

        :param context:
        :param addr_from:
        :param addr_to:
        :param func_name:
        :param amount:
        :param arg_params:
        :param kw_params:

        :return:
        """

        IconScoreContextUtil.validate_score_blacklist(context, addr_to)

        if len(context.msg_stack) == MAX_CALL_STACK_SIZE:
            raise StackOverflowException('Max call stack size exceeded')

        context.msg_stack.append(context.msg)

        prev_func_type = context.func_type
        context.current_address = addr_to
        context.msg = Message(sender=addr_from, value=amount)

        try:
            icon_score = IconScoreContextUtil.get_icon_score(context, addr_to)
            context.set_func_type_by_icon_score(icon_score, func_name)
            score_func = getattr(icon_score, ATTR_SCORE_CALL)

            if context.revision >= Revision.SCORE_FUNC_PARAMS_CHECK.value:
                metadata: ScoreElementMetadata = get_score_element_metadata(icon_score, func_name)
                verify_internal_call_arguments(metadata.signature, arg_params, kw_params)

            return score_func(func_name=func_name, arg_params=arg_params, kw_params=kw_params)
        finally:
            context.func_type = prev_func_type
            context.current_address = addr_from
            context.msg = context.msg_stack.pop()

    @staticmethod
    def emit_event_log_for_icx_transfer(context: 'IconScoreContext',
                                        from_: 'Address',
                                        to: 'Address',
                                        value: int) -> None:
        event_signature = ICX_TRANSFER_EVENT_LOG
        arguments = [from_, to, value]
        indexed_args_count = 3
        EventLogEmitter.emit_event_log(context=context,
                                       score_address=from_,
                                       event_signature=event_signature,
                                       arguments=arguments,
                                       indexed_args_count=indexed_args_count)

    @staticmethod
    def enter_call(context: 'IconScoreContext') -> None:
        """Start to call external function provided by other SCORE
        """
        if context.type != IconScoreContextType.INVOKE:
            return

        context.tx_batch.enter_call()

        context.event_log_stack.append(context.event_logs)
        context.event_logs = []

    @staticmethod
    def revert_call(context: 'IconScoreContext') -> None:
        """An exception happens during calling an external function provided by other SCORE
        Revert the states changed by this function call
        """
        if context.type != IconScoreContextType.INVOKE:
            return

        context.tx_batch.revert_call()
        context.event_logs.clear()

    @staticmethod
    def leave_call(context: 'IconScoreContext') -> None:
        """Finish to call external function provided by other SCORE
        """
        if context.type != IconScoreContextType.INVOKE:
            return

        context.tx_batch.leave_call()

        prev_event_logs = context.event_log_stack.pop()
        context.event_logs = prev_event_logs + context.event_logs



# -*- coding: utf-8 -*-
# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
from typing import TYPE_CHECKING, Iterable, Optional, Tuple

from iconcommons.logger import Logger
from .backup_manager import get_backup_filename
from ..base.exception import InvalidParamsException, InternalServiceErrorException
from ..database.db import KeyValueDatabase
from ..database.wal import WriteAheadLogReader, WALDBType
from ..icon_constant import ROLLBACK_LOG_TAG
from ..iiss.reward_calc import RewardCalcStorage
from ..iiss.reward_calc.msg_data import make_block_produce_info_key

if TYPE_CHECKING:
    from iconservice.database.db import KeyValueDatabase


TAG = ROLLBACK_LOG_TAG


class RollbackManager(object):
    """Rollback the current state to the one block previous one with a backup file

    Assume that the rollback of Reward Calculator has been already done
    Related databases: state_db, iiss_db
    """

    def __init__(self, backup_root_path: str, rc_data_path: str, state_db: 'KeyValueDatabase'):
        self._backup_root_path = backup_root_path
        self._rc_data_path = rc_data_path
        self._state_db = state_db

    def run(self, last_block_height: int, rollback_block_height: int, term_start_block_height: int):
        """Rollback to the previous block state

        :param last_block_height: the last confirmed block height
        :param rollback_block_height: the height of block to rollback to
        :param term_start_block_height: the start block height of the current term
        """
        Logger.info(tag=TAG, msg=f"run() start: "
                                 f"last_block_height={last_block_height} "
                                 f"rollback_block_height={rollback_block_height} "
                                 f"term_start_block_height={term_start_block_height}")

        self._validate_block_heights(last_block_height, rollback_block_height, term_start_block_height)

        term_change_exists = \
            self._term_change_exists(last_block_height, rollback_block_height, term_start_block_height)
        calc_end_block_height = term_start_block_height - 1
        reader = WriteAheadLogReader()
        state_db_batch = {}
        iiss_db_batch = {}

        for block_height in range(last_block_height - 1, rollback_block_height - 1, -1):
            # Make backup file with a given block_height
            path: str = self._get_backup_file_path(block_height)
            if not os.path.isfile(path):
                raise InternalServiceErrorException(f"Backup file not found: {path}")

            reader.open(path)

            # Merge backup data into state_db_batch
            self._write_batch(reader.get_iterator(WALDBType.STATE.value), state_db_batch)

            # Merge backup data into iiss_db_batch
            if not (term_change_exists and block_height > calc_end_block_height):
                self._write_batch(reader.get_iterator(WALDBType.RC.value), iiss_db_batch)

            reader.close()

        # If a term change is detected during rollback, handle the exceptions below
        if term_change_exists:
            self._remove_block_produce_info(iiss_db_batch, calc_end_block_height)
            self._rename_iiss_db_to_current_db(calc_end_block_height)

        # Commit write_batch to db
        self._commit_batch(state_db_batch, self._state_db)
        iiss_db = RewardCalcStorage.create_current_db(self._rc_data_path)
        self._commit_batch(iiss_db_batch, iiss_db)
        iiss_db.close()

        Logger.info(tag=TAG, msg="run() end")

    @staticmethod
    def _validate_block_heights(last_block_height: int, rollback_block_height: int, term_start_block_height: int):
        if last_block_height < 0:
            raise InvalidParamsException(f"Invalid lastBlockHeight: {last_block_height}")

        if rollback_block_height < 0:
            raise InvalidParamsException(f"Invalid rollbackBlockHeight: {rollback_block_height}")

        if term_start_block_height < 0:
            raise InvalidParamsException(f"Invalid termStartBlockHeight: {term_start_block_height}")

        if rollback_block_height >= last_block_height:
            raise InvalidParamsException(
                f"lastBlockHeight({last_block_height}) <= rollbackBlockHeight({rollback_block_height}")
        
    @staticmethod        
    def _term_change_exists(last_block_height: int, rollback_block_height: int, term_start_block_height: int) -> bool:
        return rollback_block_height < term_start_block_height <= last_block_height

    @staticmethod
    def _write_batch(it: Iterable[Tuple[bytes, Optional[bytes]]], batch: dict):
        for key, value in it:
            batch[key] = value

    @staticmethod
    def _commit_batch(batch: dict, db: 'KeyValueDatabase'):
        db.write_batch(batch.items())

    def _get_backup_file_path(self, block_height: int) -> str:
        """

        :param block_height: the height of block to rollback to
        :return: backup state file path
        """
        assert block_height >= 0

        filename = get_backup_filename(block_height)
        return os.path.join(self._backup_root_path, filename)

    @classmethod
    def _remove_backup_file(cls, path: str):
        try:
            os.remove(path)
        except FileNotFoundError:
            pass
        except BaseException as e:
            Logger.debug(tag=TAG, msg=str(e))

    def _rename_iiss_db_to_current_db(self, calc_end_block_height: int):
        """Rename iiss_db to current_db

        """
        Logger.debug(tag=TAG,
                     msg=f"_rename_iiss_db_to_current_db() start: calc_end_block_height={calc_end_block_height}")

        filename = RewardCalcStorage.get_iiss_rc_db_name(calc_end_block_height)
        src_path = os.path.join(self._rc_data_path, filename)
        dst_path = os.path.join(self._rc_data_path, RewardCalcStorage.CURRENT_IISS_DB_NAME)
        Logger.info(tag=TAG, msg=f"rename_iiss_db: src_path={src_path} dst_path={dst_path}")

        # Consider the case that renaming iiss_db to current_db has been already done
        if os.path.isdir(src_path):
            # Remove a new current_db
            shutil.rmtree(dst_path, ignore_errors=True)
            # Rename iiss_rc_db_{BH} to current_db
            shutil.move(src_path, dst_path)

        Logger.debug(tag=TAG, msg="_rename_iiss_db_to_current_db() end")

    @classmethod
    def _remove_block_produce_info(cls, iiss_db_batch: dict, block_height: int):
        """Remove block_produce_info of calc_period_end_block from current_db

        :param iiss_db_batch:
        :param block_height: the end block of the previous term
        :return:
        """
        Logger.debug(tag=TAG,
                     msg=f"_remove_block_produce_info() start: block_height={block_height}")

        # Remove the end calc block from iiss_db
        key: bytes = make_block_produce_info_key(block_height)
        iiss_db_batch[key] = None

        Logger.debug(tag=TAG, msg="_remove_block_produce_info() end")



# -*- coding: utf-8 -*-
# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from typing import Union, Iterable, List, Optional


class Sortable(metaclass=ABCMeta):
    @abstractmethod
    def order(self):
        pass


class SortedList(object):
    def __init__(self, sorted_list: 'SortedList' = ()):
        self._items = list(sorted_list)

    def add(self, new_item: 'Sortable'):
        """

        :param new_item:
        :return:
        """
        index = 0
        order = new_item.order()

        for item in self._items:
            if order < item.order():
                break

            index += 1

        self._items.insert(index, new_item)

    def get(self, index: int) -> Optional['Sortable']:
        try:
            return self._items[index]
        except IndexError:
            return None

    def extend(self, iterable: Iterable['Sortable']):
        self._items.extend(iterable)

    def reorder(self, item: 'Sortable'):
        self._items.remove(item)
        self.add(item)

    def index(self, item: 'Sortable') -> int:
        order = item.order()

        left: int = 0
        right: int = len(self._items)

        while left < right:
            i = (left + right) // 2
            item_in_list: 'Sortable' = self._items[i]

            if order == item_in_list.order():
                return self._precise_index(i, item)

            if order < item_in_list.order():
                right = i
            else:
                left = i + 1

        return -1

    def _precise_index(self, base_index: int, target_item: 'Sortable') -> int:
        if id(target_item) == id(self._items[base_index]):
            return base_index

        # Check the previous items
        for i in range(base_index - 1, -1, -1):
            item: 'Sortable' = self._items[i]
            if target_item.order() != item.order():
                break

            if id(target_item) == id(item):
                return i

        # Check the next items
        for i in range(base_index + 1, len(self._items)):
            item: 'Sortable' = self._items[i]
            if target_item.order() != item.order():
                break

            if id(target_item) == id(item):
                return i

        return -1

    def remove(self, item: 'Sortable') -> 'Sortable':
        index: int = self.index(item)

        if index > -1:
            return self.pop(index)

        raise ValueError(f"Value not found")

    def pop(self, index: int) -> 'Sortable':
        return self._items.pop(index)

    def append(self, item: 'Sortable'):
        """Add an item to the end

        :param item:
        :return:
        """
        size: int = len(self._items)

        # prev_item.order() should be not more than item.order()
        if size > 0:
            prev_item = self._items[size - 1]
            if item.order() < prev_item.order():
                raise ValueError("Out of order")

        self._items.append(item)

    def __iter__(self):
        for item in self._items:
            yield item

    def __getitem__(self, k: Union[int, slice]) -> Union['Sortable', List['Sortable']]:
        if isinstance(k, slice):
            return self._items[k]
        return self._items[k]

    def __setitem__(self, index: int, item: 'Sortable'):
        index = self._to_positive_index(index)

        # prev_item.order() should be not more than item.order()
        if index > 0:
            prev_item = self._items[index - 1]
            if item.order() < prev_item.order():
                raise ValueError("Out of order")

        # next_item.order() should be not less than item.order()
        if index < len(self._items) - 1:
            next_item = self._items[index + 1]
            if item.order() > next_item.order():
                raise ValueError("Out of order")

        self._items[index] = item

    def __len__(self) -> int:
        return len(self._items)

    def _to_positive_index(self, index: int) -> int:
        if index < 0:
            index += len(self._items)

        if index < 0:
            raise IndexError("Index out of range")

        return index



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from copy import deepcopy
from typing import Union, Any, get_type_hints

from .address import Address, MalformedAddress, is_icon_address_valid
from .exception import InvalidParamsException
from .type_converter_templates import ParamType, \
    type_convert_templates, ValueType, KEY_CONVERTER, CONVERT_USING_SWITCH_KEY, SWITCH_KEY
from ..icon_constant import HASH_TYPE_TABLE
from ..utils import get_main_type_from_annotations_type

score_base_support_type = (int, str, bytes, bool, Address)


class TypeConverter:
    @staticmethod
    def convert(params: Union[list, dict], param_type: ParamType) -> Any:
        if param_type is None:
            return params

        copied_params = deepcopy(params)  # to avoid corrupting original data
        converted_params = TypeConverter._convert(copied_params, type_convert_templates[param_type])
        return converted_params

    @staticmethod
    def _convert(params: Union[str, list, dict, None], template: Union[list, dict, ValueType]) -> Any:
        if TypeConverter._skip_params(params, template):
            return params

        if isinstance(template, dict) and KEY_CONVERTER in template:
            params = TypeConverter._convert_key(params, template[KEY_CONVERTER])

        if isinstance(params, dict) and isinstance(template, dict):
            new_params = {}
            for key, value in params.items():
                if TypeConverter._check_convert_using_method(key, template):
                    ref_key_table = deepcopy(new_params)
                    target_template = TypeConverter._get_convert_using_method_template(key, template)
                    new_value = TypeConverter._convert_using_switch(value, ref_key_table, target_template)
                else:
                    new_value = TypeConverter._convert(value, template.get(key))
                new_params[key] = new_value
        elif isinstance(params, list) and isinstance(template, list):
            new_params = []
            for item in params:
                if isinstance(item, list):
                    new_item = []
                    for element, templete_type in zip(item, template[0]):
                        new_element = TypeConverter._convert(element, templete_type)
                        new_item.append(new_element)
                    new_params.append(new_item)
                else:
                    new_item = TypeConverter._convert(item, template[0])
                    new_params.append(new_item)
        elif isinstance(template, ValueType):
            new_params = TypeConverter._convert_value(params, template)
        else:
            new_params = params

        return new_params

    @staticmethod
    def _convert_key(params, key_convert_dict):
        new_params = {}
        for key in params:
            if key in key_convert_dict:
                old_key = key
                new_key = key_convert_dict[old_key]
                new_params[new_key] = params[old_key]
            else:
                new_params[key] = params[key]

        return new_params

    @staticmethod
    def _check_convert_using_method(param: str, template: dict) -> bool:
        params = template.get(param)
        if isinstance(params, dict):
            return CONVERT_USING_SWITCH_KEY in params
        return False

    @staticmethod
    def _get_convert_using_method_template(key: str, template: dict) -> dict:
        tmp_params = template.get(key)
        return tmp_params.get(CONVERT_USING_SWITCH_KEY)

    @staticmethod
    def _skip_params(params: Union[str, dict, None], template: Union[list, dict, ValueType]) -> bool:
        if params is None:
            raise InvalidParamsException(f'TypeConvert Exception None value, template: {str(template)}')
        if isinstance(params, str):
            if params != "" and not template:
                return True
        elif not params or not template:
            return True
        return False
    
    @staticmethod
    def _convert_using_switch(params: Union[str, dict, None],
                              tmp_params: dict,
                              template: Union[list, dict, ValueType]) -> Any:
        if TypeConverter._skip_params(params, template):
            return params

        switch_key = template.get(SWITCH_KEY)
        template_key = tmp_params.get(switch_key)
        target_template = template.get(template_key)

        if isinstance(params, dict) and isinstance(target_template, dict):
            new_params = {}
            for key, value in params.items():
                new_value = TypeConverter._convert(value, target_template.get(key))
                new_params[key] = new_value
        elif isinstance(params, list) and isinstance(target_template, list):
            new_params = []
            for item in params:
                new_item = TypeConverter._convert(item, target_template[0])
                new_params.append(new_item)
        elif isinstance(target_template, ValueType):
            new_params = TypeConverter._convert_value(params, target_template)
        else:
            new_params = params

        return new_params

    @staticmethod
    def _convert_value(value: Any, value_type: ValueType) -> Any:
        if value_type == ValueType.INT:
            converted_value = TypeConverter._convert_value_int(value)
        elif value_type == ValueType.HEXADECIMAL:
            converted_value = TypeConverter._convert_value_hexadecimal(value)
        elif value_type == ValueType.STRING:
            converted_value = TypeConverter._convert_value_string(value)
        elif value_type == ValueType.BOOL:
            converted_value = TypeConverter._convert_value_bool(value)
        elif value_type == ValueType.ADDRESS:
            if len(value) == 0:
                converted_value = None
            else:
                converted_value = TypeConverter._convert_value_address(value)
        elif value_type == ValueType.ADDRESS_OR_MALFORMED_ADDRESS:
            converted_value = TypeConverter._convert_value_address_or_malformed_address(value)
        elif value_type == ValueType.BYTES:  # hash...(block_hash, tx_hash)
            converted_value = TypeConverter._convert_value_bytes(value)
        else:
            converted_value = value
        return converted_value

    @staticmethod
    def _convert_value_int(value: str) -> int:
        if isinstance(value, str):
            if value.startswith('0x') or value.startswith('-0x'):
                return int(value, 16)
            else:
                return int(value)
        else:
            raise InvalidParamsException(f'TypeConvert Exception int value :{value}, type: {type(value)}')

    @staticmethod
    def _convert_value_hexadecimal(value: str) -> int:
        """Convert value into integer, assuming that value is a hexadecimal string

        :param value: hexadecimal string
        :return: int
        """
        if not isinstance(value, str):
            raise InvalidParamsException(
                f'TypeConvert Exception int value :{value}, type: {type(value)}')

        return int(value, 16)

    @staticmethod
    def _convert_value_string(value: str) -> str:
        if isinstance(value, str):
            return value
        else:
            raise InvalidParamsException(f'TypeConvert Exception str value :{value}, type: {type(value)}')

    @staticmethod
    def _convert_value_bool(value: str) -> bool:
        if isinstance(value, str):
            return bool(TypeConverter._convert_value_int(value))
        else:
            raise InvalidParamsException(f'TypeConvert Exception bool value :{value}, type: {type(value)}')

    @staticmethod
    def _convert_value_address(value: str) -> 'Address':
        if isinstance(value, str):
            return Address.from_string(value)
        else:
            raise InvalidParamsException(f'TypeConvert Exception address value :{value}, type: {type(value)}')

    @staticmethod
    def _convert_value_address_or_malformed_address(value: str) -> 'Address':
        if not isinstance(value, str):
            raise InvalidParamsException(
                f'TypeConvert Exception address value :{value}, type: {type(value)}')

        if is_icon_address_valid(value):
            return Address.from_string(value)

        # This code is just used to support a legacy bug
        # Do not use MalformedAddress elsewhere
        return MalformedAddress.from_string(value)

    @staticmethod
    def _convert_value_bytes(value: str) -> bytes:
        if isinstance(value, str):
            if value.startswith('0x'):
                return bytes.fromhex(value[2:])
            else:
                return bytes.fromhex(value)
        else:
            raise InvalidParamsException(f'TypeConvert Exception bytes value :{value}, type: {type(value)}')

    @staticmethod
    def get_default_args(func):
        signature = inspect.signature(func)
        return {
            k: v.default
            for k, v in signature.parameters.items()
            if v.default is not inspect.Parameter.empty
        }

    @staticmethod
    def make_annotations_from_method(func: callable) -> dict:
        # in python 3.7, get_type_hints method return _GenericAlias type object
        # (when parameter has 'NoneType' as a default)

        hints = get_type_hints(func)
        if hints.get('return') is not None:
            del hints['return']
        return hints

    @staticmethod
    def convert_data_params(annotations: dict, kw_params: dict) -> None:
        for param_name, param_type in annotations.items():
            if param_name == "self" or param_name == "cls":
                continue

            kw_param = kw_params.get(param_name)
            if kw_param is None:
                continue

            param_type = get_main_type_from_annotations_type(param_type)
            kw_param = TypeConverter._convert_data_value(param_type, kw_param)
            kw_params[param_name] = kw_param

    @staticmethod
    def adjust_params_to_method(func: callable, kw_params: dict, remove_invalid_param: bool = False):
        hints = TypeConverter.make_annotations_from_method(func)
        default_args = TypeConverter.get_default_args(func)

        # check user input argument name is valid
        invalid_keys = []
        for key in kw_params.keys():
            try:
                _type = hints[key]
            except KeyError:
                invalid_keys.append(key)

        if len(invalid_keys) > 0:
            if remove_invalid_param:
                for key in invalid_keys:
                    del kw_params[key]
                if len(kw_params) == 0:
                    # input invalid params only
                    raise InvalidParamsException(f"There is no valid parameters")
            else:
                raise InvalidParamsException(f"Invalid parameters '{invalid_keys}'")

        # check required argument is exist in user input
        for param_name, param_type in hints.items():
            if param_name == "self" or param_name == "cls":
                continue

            try:
                param = kw_params[param_name]
            except KeyError:
                # has no input for this parameter
                try:
                    default_args[param_name]
                except KeyError:
                    # has no default value
                    raise InvalidParamsException(f"There is no '{param_name}' parameter")
                # has default value. pass type converting
                continue

            # all type can have None value
            if param is None:
                continue

            param_type = get_main_type_from_annotations_type(param_type)
            param = TypeConverter._convert_data_value(param_type, param)
            kw_params[param_name] = param

    @staticmethod
    def _convert_data_value(annotation_type: type, param: Any) -> Any:
        if annotation_type == int:
            param = TypeConverter._convert_value_int(param)
        elif annotation_type == str:
            param = TypeConverter._convert_value_string(param)
        elif annotation_type == bool:
            param = TypeConverter._convert_value_bool(param)
        elif annotation_type == Address:
            param = TypeConverter._convert_value_address(param)
        elif annotation_type == bytes:
            param = TypeConverter._convert_value_bytes(param)
        return param

    @staticmethod
    def convert_type_reverse(value: Any):
        if isinstance(value, dict):
            for k, v in value.items():
                if isinstance(v, bytes):
                    is_hash = k in HASH_TYPE_TABLE
                    value[k] = TypeConverter._convert_bytes_reverse(v, is_hash)
                else:
                    value[k] = TypeConverter.convert_type_reverse(v)
        elif isinstance(value, list):
            for i, v in enumerate(value):
                value[i] = TypeConverter.convert_type_reverse(v)
        elif isinstance(value, int):
            value = hex(value)
        elif isinstance(value, Address):
            value = str(value)
        elif isinstance(value, bytes):
            value = TypeConverter._convert_bytes_reverse(value)
        return value

    @staticmethod
    def _convert_bytes_reverse(value: bytes, is_hash: bool = False):
        if is_hash:
            # if the value is of 'txHash' or 'blockHash', excludes '0x' prefix
            return bytes.hex(value)
        else:
            return f'0x{bytes.hex(value)}'



# -*- coding: utf-8 -*-

# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Tuple, List

from ..base.ComponentBase import StorageBase
from ..database.db import MetaContextDatabase, ContextDatabase
from ..prep.prep_address_converter import PRepAddressConverter
from ..utils.msgpack_for_db import MsgPackForDB

if TYPE_CHECKING:
    from ..iconscore.icon_score_context import IconScoreContext
    from ..base.address import Address


class Storage(StorageBase):
    _KEY_LAST_CALC_INFO = b'last_calc_info'
    _KEY_LAST_TERM_INFO = b'last_term_info'
    _KEY_LAST_MAIN_PREPS = b'last_main_preps'

    _KEY_PREV_NODE_ADDRESS_MAPPER = b'prev_node_address_mapper'

    def __init__(self, db: 'ContextDatabase'):
        super().__init__(MetaContextDatabase(db.key_value_db))

    def put_last_calc_info(self,
                           context: 'IconScoreContext',
                           start: int,
                           end: int):
        version = 0
        value: bytes = MsgPackForDB.dumps([version, start, end])
        self._db.put(context, self._KEY_LAST_CALC_INFO, value)

    def get_last_calc_info(self, context: 'IconScoreContext') -> Tuple[int, int]:
        value: bytes = self._db.get(context, self._KEY_LAST_CALC_INFO)
        if value is None:
            return -1, -1
        data: list = MsgPackForDB.loads(value)
        _version = data[0]
        return data[1], data[2]

    def put_last_term_info(self,
                           context: 'IconScoreContext',
                           start: int,
                           end: int):
        version = 0
        value: bytes = MsgPackForDB.dumps([version, start, end])
        self._db.put(context, self._KEY_LAST_TERM_INFO, value)

    def get_last_term_info(self, context: 'IconScoreContext') -> Tuple[int, int]:
        value: bytes = self._db.get(context, self._KEY_LAST_TERM_INFO)
        if value is None:
            return -1, -1
        data: list = MsgPackForDB.loads(value)
        _version = data[0]
        return data[1], data[2]

    def put_last_main_preps(self,
                            context: 'IconScoreContext',
                            main_preps: List['Address']):
        version = 0
        value: bytes = MsgPackForDB.dumps([version, main_preps])
        self._db.put(context, self._KEY_LAST_MAIN_PREPS, value)

    def get_last_main_preps(self, context: 'IconScoreContext') -> List['Address']:
        value: bytes = self._db.get(context, self._KEY_LAST_MAIN_PREPS)
        if value is None:
            return []
        data: list = MsgPackForDB.loads(value)
        _version = data[0]
        return data[1]

    def put_prep_address_converter(self,
                                   context: 'IconScoreContext',
                                   prep_address_converter: 'PRepAddressConverter'):
        value: bytes = prep_address_converter.to_bytes()
        self._db.put(context, self._KEY_PREV_NODE_ADDRESS_MAPPER, value)

    def get_prep_address_converter(self, context: 'IconScoreContext') -> 'PRepAddressConverter':
        value: bytes = self._db.get(context, self._KEY_PREV_NODE_ADDRESS_MAPPER)
        return PRepAddressConverter.from_bytes(value)



# -*- coding: utf-8 -*-

# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import hashlib
from typing import Union, Iterable

from .hash_origin_generator import HashOriginGeneratorV1
from .merkle_tree import MerkleTree


class HashGenerator:
    _SALT = "icx_sendTransaction"
    _ORIGIN_GENERATOR = HashOriginGeneratorV1()

    @classmethod
    def generate_origin(cls, origin_data: dict) -> str:
        copied_origin_data = copy.deepcopy(origin_data)
        return cls._ORIGIN_GENERATOR.generate(copied_origin_data)

    @classmethod
    def generate_salted_origin(cls, origin_data: dict) -> str:
        def _gen():
            if HashGenerator._SALT is not None:
                yield HashGenerator._SALT
            yield cls.generate_origin(origin_data)
        return '.'.join(_gen())

    @classmethod
    def generate_hash(cls, origin_data: dict) -> bytes:
        origin = cls.generate_salted_origin(origin_data)
        return hashlib.sha3_256(origin.encode()).digest()


class RootHashGenerator:
    @classmethod
    def generate_root_hash(cls, values: Union[Iterable, bytes, bytearray], do_hash=False) -> bytes:
        merkle_tree = MerkleTree()
        merkle_tree.add_leaf(values, do_hash)
        merkle_tree.make_tree()
        return merkle_tree.get_merkle_root()



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import IntEnum

from ..icon_constant import IssueDataKey


class ParamType(IntEnum):
    BLOCK = 0

    INVOKE_TRANSACTION = 100
    ACCOUNT_DATA = 101
    CALL_DATA = 102
    DEPLOY_DATA = 103
    TRANSACTION_PARAMS_DATA = 104
    DEPOSIT_DATA = 105
    BASE_DATA = 106

    INVOKE = 200

    QUERY = 300
    ICX_CALL = 301
    ICX_GET_BALANCE = 302
    ICX_GET_TOTAL_SUPPLY = 303
    ICX_GET_SCORE_API = 304
    ISE_GET_STATUS = 305

    WRITE_PRECOMMIT = 400
    # REMOVE_PRECOMMIT = 500
    ROLLBACK = 501

    VALIDATE_TRANSACTION = 600

    # IISS
    IISS = 700
    IISS_SET_STAKE = 701
    IISS_GET_STAKE = 702
    IISS_SET_DELEGATION = 703
    IISS_GET_DELEGATION = 704
    IISS_CLAIM_ISCORE = 705
    IISS_QUERY_ISCORE = 706
    IISS_REG_PREP = 707
    IISS_UNREG_PREP = 708
    IISS_SET_PREP = 709
    IISS_GET_PREP = 710
    IISS_GET_PREP_DELEGATION_INFO = 711
    IISS_GET_MAIN_PREP_LIST = 712
    IISS_GET_PREP_LIST = 713
    IISS_SET_GOVERNANCE_VARIABLES = 714


class ValueType(IntEnum):
    IGNORE = 0
    LATER = 1
    INT = 2
    STRING = 3
    BOOL = 4
    ADDRESS = 5
    BYTES = 6

    # For backward compatibility (TestNet)
    ADDRESS_OR_MALFORMED_ADDRESS = 7
    HEXADECIMAL = 8


type_convert_templates = {}
CONVERT_USING_SWITCH_KEY = 'CONVERT_USING_SWITCH_KEY'
SWITCH_KEY = "SWITCH_KEY"
KEY_CONVERTER = 'KEY_CONVERTER'


class ConstantKeys:
    BLOCK_HEIGHT = "blockHeight"
    BLOCK_HASH = "blockHash"
    OLD_BLOCK_HASH = "oldBlockHash"
    NEW_BLOCK_HASH = "newBlockHash"
    TIMESTAMP = "timestamp"
    PREV_BLOCK_HASH = "prevBlockHash"

    NAME = "name"
    ADDRESS = "address"
    BALANCE = "balance"

    METHOD = "method"
    PARAMS = "params"

    CONTENT_TYPE = "contentType"
    CONTENT = "content"

    TX_HASH = "txHash"
    VERSION = "version"
    FROM = "from"
    TO = "to"
    VALUE = "value"
    STEP_LIMIT = "stepLimit"
    FEE = "fee"
    NONCE = "nonce"
    SIGNATURE = "signature"

    DATA_TYPE = "dataType"
    DATA = "data"
    CALL = "call"
    DEPLOY = "deploy"

    BASE = "base"
    PREP = IssueDataKey.PREP
    PREP_INCENTIVE = IssueDataKey.IREP
    PREP_REWARD_RATE = IssueDataKey.RREP
    PREP_TOTAL_DELEGATION = IssueDataKey.TOTAL_DELEGATION
    PREP_VALUE = IssueDataKey.VALUE

    EEP = IssueDataKey.EEP
    EEP_INCENTIVE = IssueDataKey.IEEP
    EEP_REWARD_RATE = IssueDataKey.REEP
    EEP_TOTAL_DELEGATION = IssueDataKey.TOTAL_DELEGATION
    EEP_VALUE = IssueDataKey.VALUE

    DAPP = IssueDataKey.DAPP
    DAPP_INCENTIVE = IssueDataKey.IDAPP
    DAPP_REWARD_RATE = IssueDataKey.RDAPP
    DAPP_TOTAL_DELEGATION = IssueDataKey.TOTAL_DELEGATION
    DAPP_VALUE = IssueDataKey.VALUE

    ISSUE_RESULT = IssueDataKey.ISSUE_RESULT
    COVERED_BY_FEE = IssueDataKey.COVERED_BY_FEE
    COVERED_BY_OVER_ISSUED_ICX = IssueDataKey.COVERED_BY_OVER_ISSUED_ICX
    ISSUE = IssueDataKey.ISSUE

    OLD_TX_HASH = "tx_hash"

    GENESIS_DATA = "genesisData"
    ACCOUNTS = "accounts"
    MESSAGE = "message"

    BLOCK = "block"
    TRANSACTIONS = "transactions"

    IS_BLOCK_EDITABLE = 'isBlockEditable'
    PREV_BLOCK_GENERATOR = "prevBlockGenerator"
    PREV_BLOCK_VALIDATORS = "prevBlockValidators"
    PREV_BLOCK_VOTES = "prevBlockVotes"

    FILTER = "filter"

    ICX_CALL = "icx_call"
    ICX_GET_BALANCE = "icx_getBalance"
    ICX_GET_TOTAL_SUPPLY = "icx_getTotalSupply"
    ICX_GET_SCORE_API = "icx_getScoreApi"
    ISE_GET_STATUS = "ise_getStatus"

    DEPOSIT_TERM = "term"
    DEPOSIT_ID = "id"

    # IISS
    DELEGATIONS = "delegations"
    COUNTRY = "country"
    CITY = "city"
    EMAIL = 'email'
    WEBSITE = 'website'
    DETAILS = 'details'
    P2P_ENDPOINT = 'p2pEndpoint'
    PREP_ID = 'id'
    START_RANKING = "startRanking"
    END_RANKING = "endRanking"
    IREP = "irep"
    IREP_BLOCK_HEIGHT = "irepUpdateBlockHeight"
    NODE_ADDRESS = "nodeAddress"


type_convert_templates[ParamType.BLOCK] = {
    ConstantKeys.BLOCK_HEIGHT: ValueType.INT,
    ConstantKeys.BLOCK_HASH: ValueType.BYTES,
    ConstantKeys.TIMESTAMP: ValueType.INT,
    ConstantKeys.PREV_BLOCK_HASH: ValueType.BYTES,
}

type_convert_templates[ParamType.ACCOUNT_DATA] = {
    ConstantKeys.NAME: ValueType.STRING,
    ConstantKeys.ADDRESS: ValueType.ADDRESS,
    ConstantKeys.BALANCE: ValueType.INT
}

type_convert_templates[ParamType.CALL_DATA] = {
    ConstantKeys.METHOD: ValueType.STRING,
    ConstantKeys.PARAMS: ValueType.LATER
}

type_convert_templates[ParamType.DEPLOY_DATA] = {
    ConstantKeys.CONTENT_TYPE: ValueType.STRING,
    ConstantKeys.CONTENT: ValueType.IGNORE,
    ConstantKeys.PARAMS: ValueType.LATER
}

type_convert_templates[ParamType.BASE_DATA] = {
    ConstantKeys.PREP: {
        ConstantKeys.PREP_INCENTIVE: ValueType.INT,
        ConstantKeys.PREP_REWARD_RATE: ValueType.INT,
        ConstantKeys.PREP_TOTAL_DELEGATION: ValueType.INT,
        ConstantKeys.PREP_VALUE: ValueType.INT
    },
    ConstantKeys.EEP: {
        ConstantKeys.EEP_INCENTIVE: ValueType.INT,
        ConstantKeys.EEP_REWARD_RATE: ValueType.INT,
        ConstantKeys.EEP_TOTAL_DELEGATION: ValueType.INT,
        ConstantKeys.EEP_VALUE: ValueType.INT
    },
    ConstantKeys.DAPP: {
        ConstantKeys.DAPP_INCENTIVE: ValueType.INT,
        ConstantKeys.DAPP_REWARD_RATE: ValueType.INT,
        ConstantKeys.DAPP_TOTAL_DELEGATION: ValueType.INT,
        ConstantKeys.DAPP_VALUE: ValueType.INT
    },
    ConstantKeys.ISSUE_RESULT: {
        ConstantKeys.COVERED_BY_FEE: ValueType.INT,
        ConstantKeys.COVERED_BY_OVER_ISSUED_ICX: ValueType.INT,
        ConstantKeys.ISSUE: ValueType.INT
    }

}

type_convert_templates[ParamType.TRANSACTION_PARAMS_DATA] = {
    ConstantKeys.VERSION: ValueType.INT,
    ConstantKeys.TX_HASH: ValueType.BYTES,
    ConstantKeys.FROM: ValueType.ADDRESS,
    ConstantKeys.TO: ValueType.ADDRESS_OR_MALFORMED_ADDRESS,
    ConstantKeys.VALUE: ValueType.HEXADECIMAL,
    ConstantKeys.STEP_LIMIT: ValueType.INT,
    ConstantKeys.FEE: ValueType.HEXADECIMAL,
    ConstantKeys.TIMESTAMP: ValueType.INT,
    ConstantKeys.NONCE: ValueType.INT,
    ConstantKeys.SIGNATURE: ValueType.IGNORE,
    ConstantKeys.DATA_TYPE: ValueType.STRING,
    ConstantKeys.DATA: {
        CONVERT_USING_SWITCH_KEY: {
            SWITCH_KEY: ConstantKeys.DATA_TYPE,
            ConstantKeys.CALL: type_convert_templates[ParamType.CALL_DATA],
            ConstantKeys.DEPLOY: type_convert_templates[ParamType.DEPLOY_DATA],
            ConstantKeys.BASE: type_convert_templates[ParamType.BASE_DATA]
        }
    },
    KEY_CONVERTER: {
        ConstantKeys.OLD_TX_HASH: ConstantKeys.TX_HASH
    }
}

type_convert_templates[ParamType.INVOKE_TRANSACTION] = {
    ConstantKeys.METHOD: ValueType.STRING,
    ConstantKeys.PARAMS: type_convert_templates[ParamType.TRANSACTION_PARAMS_DATA],
    ConstantKeys.GENESIS_DATA: {
        ConstantKeys.ACCOUNTS: [
            type_convert_templates[ParamType.ACCOUNT_DATA]
        ],
        ConstantKeys.MESSAGE: ValueType.STRING
    }
}

type_convert_templates[ParamType.INVOKE] = {
    ConstantKeys.BLOCK: type_convert_templates[ParamType.BLOCK],
    ConstantKeys.TRANSACTIONS: [
        type_convert_templates[ParamType.INVOKE_TRANSACTION]
    ],
    ConstantKeys.IS_BLOCK_EDITABLE: ValueType.BOOL,
    ConstantKeys.PREV_BLOCK_GENERATOR: ValueType.ADDRESS,
    ConstantKeys.PREV_BLOCK_VALIDATORS: [ValueType.ADDRESS],
    ConstantKeys.PREV_BLOCK_VOTES: [
        [
            ValueType.ADDRESS,
            ValueType.INT
        ]
    ]
}

type_convert_templates[ParamType.ICX_CALL] = {
    ConstantKeys.VERSION: ValueType.INT,
    ConstantKeys.FROM: ValueType.ADDRESS,
    ConstantKeys.TO: ValueType.ADDRESS,
    ConstantKeys.DATA_TYPE: ValueType.STRING,
    ConstantKeys.DATA: ValueType.LATER
}
type_convert_templates[ParamType.ICX_GET_BALANCE] = {
    ConstantKeys.VERSION: ValueType.INT,
    ConstantKeys.ADDRESS: ValueType.ADDRESS_OR_MALFORMED_ADDRESS
}
type_convert_templates[ParamType.ICX_GET_TOTAL_SUPPLY] = {
    ConstantKeys.VERSION: ValueType.INT
}
type_convert_templates[ParamType.ICX_GET_SCORE_API] = type_convert_templates[ParamType.ICX_GET_BALANCE]

type_convert_templates[ParamType.ISE_GET_STATUS] = {
    ConstantKeys.FILTER: [ValueType.STRING]
}

type_convert_templates[ParamType.QUERY] = {
    ConstantKeys.METHOD: ValueType.STRING,
    ConstantKeys.PARAMS: {
        CONVERT_USING_SWITCH_KEY: {
            SWITCH_KEY: ConstantKeys.METHOD,
            ConstantKeys.ICX_CALL: type_convert_templates[ParamType.ICX_CALL],
            ConstantKeys.ICX_GET_BALANCE: type_convert_templates[ParamType.ICX_GET_BALANCE],
            ConstantKeys.ICX_GET_TOTAL_SUPPLY: type_convert_templates[ParamType.ICX_GET_TOTAL_SUPPLY],
            ConstantKeys.ICX_GET_SCORE_API: type_convert_templates[ParamType.ICX_GET_SCORE_API],
            ConstantKeys.ISE_GET_STATUS: type_convert_templates[ParamType.ISE_GET_STATUS]
        }
    }
}

type_convert_templates[ParamType.WRITE_PRECOMMIT] = {
    ConstantKeys.BLOCK_HEIGHT: ValueType.INT,
    ConstantKeys.BLOCK_HASH: ValueType.BYTES,
    ConstantKeys.OLD_BLOCK_HASH: ValueType.BYTES,
    ConstantKeys.NEW_BLOCK_HASH: ValueType.BYTES
}

type_convert_templates[ParamType.ROLLBACK] = {
    ConstantKeys.BLOCK_HEIGHT: ValueType.INT,
    ConstantKeys.BLOCK_HASH: ValueType.BYTES
}

type_convert_templates[ParamType.VALIDATE_TRANSACTION] = {
    ConstantKeys.METHOD: ValueType.STRING,
    ConstantKeys.PARAMS: type_convert_templates[ParamType.TRANSACTION_PARAMS_DATA]
}

# DEPOSIT
type_convert_templates[ParamType.DEPOSIT_DATA] = {
    ConstantKeys.DEPOSIT_ID: ValueType.BYTES,
    ConstantKeys.DEPOSIT_TERM: ValueType.INT,
}

# IISS
type_convert_templates[ParamType.IISS_SET_STAKE] = {
    ConstantKeys.VALUE: ValueType.INT
}

type_convert_templates[ParamType.IISS_GET_STAKE] = {
    ConstantKeys.ADDRESS: ValueType.ADDRESS
}

type_convert_templates[ParamType.IISS_SET_DELEGATION] = [
    {
        ConstantKeys.ADDRESS: ValueType.ADDRESS,
        ConstantKeys.VALUE: ValueType.INT
    }
]

type_convert_templates[ParamType.IISS_GET_DELEGATION] = type_convert_templates[ParamType.IISS_GET_STAKE]

type_convert_templates[ParamType.IISS_CLAIM_ISCORE] = {}

type_convert_templates[ParamType.IISS_QUERY_ISCORE] = type_convert_templates[ParamType.IISS_GET_STAKE]

type_convert_templates[ParamType.IISS_REG_PREP] = {
    ConstantKeys.NAME: ValueType.STRING,
    ConstantKeys.COUNTRY: ValueType.STRING,
    ConstantKeys.CITY: ValueType.STRING,
    ConstantKeys.EMAIL: ValueType.STRING,
    ConstantKeys.WEBSITE: ValueType.STRING,
    ConstantKeys.DETAILS: ValueType.STRING,
    ConstantKeys.P2P_ENDPOINT: ValueType.STRING,
    ConstantKeys.NODE_ADDRESS: ValueType.ADDRESS,
}

type_convert_templates[ParamType.IISS_SET_PREP] = {
    ConstantKeys.NAME: ValueType.STRING,
    ConstantKeys.COUNTRY: ValueType.STRING,
    ConstantKeys.CITY: ValueType.STRING,
    ConstantKeys.EMAIL: ValueType.STRING,
    ConstantKeys.WEBSITE: ValueType.STRING,
    ConstantKeys.DETAILS: ValueType.STRING,
    ConstantKeys.P2P_ENDPOINT: ValueType.STRING,
    ConstantKeys.NODE_ADDRESS: ValueType.ADDRESS
}

type_convert_templates[ParamType.IISS_UNREG_PREP] = type_convert_templates[ParamType.IISS_CLAIM_ISCORE]

type_convert_templates[ParamType.IISS_GET_PREP] = type_convert_templates[ParamType.IISS_GET_STAKE]

type_convert_templates[ParamType.IISS_GET_PREP_DELEGATION_INFO] = \
    type_convert_templates[ParamType.IISS_GET_STAKE]

type_convert_templates[ParamType.IISS_GET_MAIN_PREP_LIST] = type_convert_templates[ParamType.IISS_CLAIM_ISCORE]

type_convert_templates[ParamType.IISS_GET_PREP_LIST] = {
    ConstantKeys.START_RANKING: ValueType.INT,
    ConstantKeys.END_RANKING: ValueType.INT
}

type_convert_templates[ParamType.IISS_SET_GOVERNANCE_VARIABLES] = {
    ConstantKeys.IREP: ValueType.INT
}



# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Optional, TYPE_CHECKING, List

from .container import Container
from .data.value import Value, VALUE_MAPPER
from ..base.ComponentBase import StorageBase
from ..database.db import ContextDatabase
from ..icon_constant import IconNetworkValueType
from ..utils.msgpack_for_db import MsgPackForDB

if TYPE_CHECKING:
    from ..iconscore.icon_score_context import IconScoreContext


class Storage(StorageBase):
    MIGRATION_FLAG: bytes = b'mf'

    def __init__(self, db: 'ContextDatabase'):
        super().__init__(db)

    def get_container(self, context: 'IconScoreContext') -> Optional['Container']:
        """
        Load container from DB after migration

        :param context:
        :return: Return'None' if migration has not been finished
        """
        is_migrated: bool = self._get_migration_flag(context)
        container: Optional['Container'] = None
        if not is_migrated:
            return container

        container: 'Container' = Container(is_migrated)
        for type_ in IconNetworkValueType:
            value: Optional['Value'] = self._get_value(context, type_)
            if value is not None:
                container.set_inv(value, is_open=True)
        return container

    def migrate(self, context: 'IconScoreContext', data: List['Value']):
        for value in data:
            self.put_value(context, value)
        self._db.put(context, self.MIGRATION_FLAG, MsgPackForDB.dumps(True))

    def _get_migration_flag(self, context: 'IconScoreContext') -> bool:
        return bool(self._db.get(context, self.MIGRATION_FLAG))

    def put_value(self, context: 'IconScoreContext', value: 'Value'):
        self._db.put(context, value.make_key(), value.to_bytes())

    def _get_value(self, context: 'IconScoreContext', type_: 'IconNetworkValueType') -> Optional['Value']:
        assert isinstance(type_, IconNetworkValueType)
        value: Optional[bytes] = self._db.get(context, Value.PREFIX + type_.value)
        if value is None:
            return None

        value: 'value' = VALUE_MAPPER[type_].from_bytes(value)
        return value



# -*- coding: utf-8 -*-
# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import Flag, auto

from .batch import BlockBatchValue
from ..icon_constant import DATA_BYTE_ORDER
from ..iiss.reward_calc.msg_data import TxData
from ..iiss.reward_calc.storage import Storage, get_rc_version
from ..utils.msgpack_for_db import MsgPackForDB

__all__ = (
    "WriteAheadLogWriter", "WriteAheadLogReader", "WALogable", "StateWAL", "IissWAL", "WALState", "WALDBType"
)

import struct
from abc import ABCMeta
from typing import Optional, Tuple, Iterable, List
import os
from enum import Enum

import msgpack
from iconcommons.logger import Logger

from ..base.block import Block
from ..base.exception import AccessDeniedException, InvalidParamsException
from ..base.exception import InternalServiceErrorException, IllegalFormatException
from ..utils import bytes_to_hex
from ..database.batch import BlockBatch, TransactionBatchValue


TAG = "WAL"
_MAGIC_KEY = b"IWAL"
_FILE_VERSION = 1
_HEADER_SIZE = 52
_HEADER_STRUCT_FORMAT = ">4sIII32sI"

# FILE OFFSET
_OFFSET_MAGIC_KEY = 0
_OFFSET_VERSION = _OFFSET_MAGIC_KEY + 4
_OFFSET_REVISION = _OFFSET_VERSION + 4
_OFFSET_STATE = _OFFSET_REVISION + 4
_OFFSET_INSTANT_BLOCK_HASH = _OFFSET_STATE + 4
_OFFSET_LOG_COUNT = _OFFSET_INSTANT_BLOCK_HASH + 32
_OFFSET_LOG_START_OFFSETS = _OFFSET_LOG_COUNT + 4


class WALDBType(Enum):
    RC = 0
    STATE = 1


class WALState(Flag):
    CALC_PERIOD_START_BLOCK = auto()
    # Write WAL to rc_db
    WRITE_RC_DB = auto()
    # Write WAL to state_db
    WRITE_STATE_DB = auto()
    # Send COMMIT_BLOCK message to rc
    SEND_COMMIT_BLOCK = auto()
    # Send CALCULATE message to rc
    SEND_CALCULATE = auto()

    # Means All flags are on
    ALL = 0xFFFFFFFF


def block_batch_value_to_bytes(block_batch_value: 'BlockBatchValue') -> Optional[bytes]:
    if not isinstance(block_batch_value, BlockBatchValue):
        raise InvalidParamsException(f"Invalid value type: {type(block_batch_value)}")
    return block_batch_value.value


class WALogable(metaclass=ABCMeta):
    def __iter__(self) -> Tuple[bytes, Optional[bytes]]:
        pass


class StateWAL(WALogable):
    def __init__(self, block_batch: 'BlockBatch', converter: Optional[callable] = block_batch_value_to_bytes):
        self.block_batch: 'BlockBatch' = block_batch
        self.converter: callable = converter

    def __iter__(self) -> Tuple[bytes, Optional[bytes]]:
        for key, value in self.block_batch.items():
            if self.converter:
                value = self.converter(value)
            yield key, value


class IissWAL(WALogable):
    def __init__(self, rc_batch: list, tx_index: int, revision: int = -1):
        self._rc_batch: list = rc_batch
        self._tx_index: int = tx_index
        # If revision is not -1, should put revision and version to rc db
        self._revision: int = revision
        self._version: int = self._get_version()

        self._final_tx_index: Optional[int] = None

    @property
    def final_tx_index(self) -> Optional[int]:
        # If None, means there were any data to put
        return self._final_tx_index

    def _get_version(self):
        version: int = -1
        if self._revision == -1:
            return version
        else:
            version: int = get_rc_version(self._revision)
            return version

    def __iter__(self) -> Tuple[bytes, Optional[bytes]]:
        tx_index = self._tx_index

        # In case of the start block of calc period, put version, revision
        if self._revision != -1:
            # todo: refactoring (same logic exists in reward_calc storage
            key: bytes = Storage.KEY_FOR_VERSION_AND_REVISION
            value: bytes = MsgPackForDB.dumps([self._version, self._revision])
            yield key, value

        for iiss_data in self._rc_batch:
            if isinstance(iiss_data, TxData):
                tx_index += 1
                key: bytes = iiss_data.make_key(tx_index)
            else:
                key: bytes = iiss_data.make_key()
            value: bytes = iiss_data.make_value()
            yield key, value

        if tx_index > self._tx_index:
            key: bytes = Storage.KEY_FOR_GETTING_LAST_TRANSACTION_INDEX
            value: bytes = tx_index.to_bytes(8, DATA_BYTE_ORDER)
            yield key, value

        self._final_tx_index = tx_index


def _uint32_to_bytes(value: int):
    return value.to_bytes(4, "big", signed=False)


def _bytes_to_uint32(data: bytes):
    return int.from_bytes(data, "big", signed=False)


class WriteAheadLogWriter(object):
    """Write write-ahead-logging for block, state_db and rc_db on commit

    | magic_key(4) | version(4) | revision(4) | state(4) | instant_block_hash(32) |
    | data count(4) | log start address_0 (4) | log start address_1 (4) | ...
    | block data size(4) | block data | size(4) | data | size(4) | data | ...

    Every number is written in big endian format
    """

    def __init__(self, revision: int, max_log_count: int, block: 'Block', instant_block_hash: bytes):
        Logger.debug(tag=TAG,
                     msg=f"__init__(revision={revision}, "
                         f"max_log_out={max_log_count}, "
                         f"block={block} start")

        self._magic_key = _MAGIC_KEY
        self._version = _FILE_VERSION
        self._revision: int = revision
        self._state: int = 0
        self._max_log_count: int = max_log_count
        self._log_count: int = 0

        self._instant_block_hash = instant_block_hash
        self._block = block
        self._fp = None

        Logger.debug(tag=TAG, msg="__init__() end")

    def open(self, path: str):
        if self._fp is not None:
            raise InternalServiceErrorException("WAL file pointer is not None")

        try:
            self._fp = open(path, "wb")
            self._write_header()
            self._write_block()
        except:
            self._fp = None
            raise

    def _write_header(self) -> int:
        values = [
            self._magic_key,
            self._version,
            self._revision,
            self._state,
            self._instant_block_hash,
            self._log_count
        ]

        for _ in range(self._max_log_count):
            values.append(0)

        struct_format = _HEADER_STRUCT_FORMAT + "I" * self._max_log_count
        data: bytes = struct.pack(struct_format, *values)
        return self._fp.write(data)

    def flush(self):
        fp = self._fp

        if fp:
            fp.flush()
            os.fsync(fp.fileno())

    def close(self):
        if self._fp:
            self._fp.close()
            self._fp = None

    def _write_block(self) -> int:
        block = self._block

        data: bytes = block.to_bytes(self._revision)
        self._write_uint32(len(data))
        ret = self._fp.write(data) + 4

        return ret

    def write_walogable(self, it: Iterable[Tuple[bytes, Optional[bytes]]]) -> int:
        self._fp.seek(0, 2)
        start_offset: int = self._fp.tell()

        size = 0
        self._write_uint32(size)

        for key, value in it:
            size += self._write_key_value(key, value)

        # Return to the WALogable start offset, writing its data size
        self._fp.seek(-size - 4, 1)
        self._write_uint32(size)

        self._write_log_start_offset(self._log_count, start_offset)
        self._write_log_count()

        return size

    def _write_key_value(self, key: bytes, value: bytes) -> int:
        assert isinstance(key, bytes)

        data: bytes = msgpack.packb([key, value])
        size = self._fp.write(data)
        assert size == len(data)

        return size

    def write_state(self, state: int, add: bool = False):
        offset = _OFFSET_STATE

        if add:
            state |= self._state

        self._fp.seek(offset, 0)
        self._write_uint32(state)
        self._state = state

    def _write_log_count(self) -> int:
        offset = _OFFSET_LOG_COUNT
        self._log_count += 1

        self._fp.seek(offset, 0)
        data: bytes = _uint32_to_bytes(self._log_count)
        return self._fp.write(data)

    def _write_log_start_offset(self, index: int, start_offset: int) -> int:
        offset = _OFFSET_LOG_START_OFFSETS + index * 4
        self._fp.seek(offset, 0)

        return self._write_uint32(start_offset)

    def _write_uint32(self, value: int) -> int:
        return self._fp.write(_uint32_to_bytes(value))

    def _check_file_pointer(self):
        if self._fp is None:
            raise AccessDeniedException("WAL not ready")


class WriteAheadLogReader(object):
    """Read data from a write ahead log file

    """

    def __init__(self):
        self._magic_key: Optional[bytes] = None
        self._version: int = 0
        self._revision: int = 0
        self._state: int = 0
        self._log_count: int = 0
        self._log_start_offsets: Optional[List[int]] = None
        self._instant_block_hash: bytes = b""
        self._block: Optional['Block'] = None

        self._fp = None

    @property
    def magic_key(self) -> Optional[bytes]:
        return self._magic_key

    @property
    def version(self) -> int:
        return self._version

    @property
    def revision(self) -> int:
        return self._revision

    @property
    def state(self) -> int:
        return self._state

    @property
    def instant_block_hash(self) -> bytes:
        return self._instant_block_hash

    @property
    def block(self) -> Optional['Block']:
        return self._block

    @property
    def log_count(self) -> int:
        return self._log_count

    def __str__(self):
        return f"version={self._version}, " \
               f"state={self._state}, " \
               f"instant_block_hash={bytes_to_hex(self._instant_block_hash)}, " \
               f"log_count={self._log_count}, " \
               f"block={self._block}"

    def open(self, path: str):
        self._fp = open(path, "rb")
        self._read_header()
        self._read_block()

    def close(self):
        if self._fp:
            self._fp.close()
            self._fp = None

    def _read_header(self):
        data: bytes = self._fp.read(_HEADER_SIZE)
        self._check_bytes_data(data, _HEADER_SIZE)

        magic_key, version, revision, state, instant_block_hash, log_count = \
            struct.unpack_from(_HEADER_STRUCT_FORMAT, data)

        if magic_key != _MAGIC_KEY:
            raise IllegalFormatException(f"Invalid magic key: {bytes_to_hex(data)}")

        if version != _FILE_VERSION:
            raise IllegalFormatException(
                f"Invalid version: Actual({version}) != Expected({_FILE_VERSION})")

        self._magic_key = magic_key
        self._version = version
        self._revision = revision
        self._state = state
        self._log_count = log_count
        self._instant_block_hash = instant_block_hash
        self._log_start_offsets = []

        for _ in range(self._log_count):
            offset = self._read_uint32()
            self._log_start_offsets.append(offset)

    def _read_block(self):
        size: int = self._read_uint32()
        data: bytes = self._fp.read(size)
        self._check_bytes_data(data, size)

        self._block = Block.from_bytes(data)

    def _read_uint32(self) -> int:
        size = 4
        data: bytes = self._fp.read(size)
        self._check_bytes_data(data, size)

        return _bytes_to_uint32(data)

    def get_iterator(self, index: int) -> Iterable[Tuple[bytes, Optional[bytes]]]:
        self._seek_to_log_start_offset(index)
        size: int = self._read_uint32()

        unpacker = msgpack.Unpacker(use_list=False, raw=True)

        while size > 0:
            size_to_read = min(size, 16384)
            data: bytes = self._fp.read(size_to_read)
            self._check_bytes_data(data, size_to_read)

            size -= size_to_read

            unpacker.feed(data)
            for key, value in unpacker:
                yield key, value

    def _seek_to_log_start_offset(self, index: int):
        offset = self._log_start_offsets[index]
        self._fp.seek(offset, 0)

    @classmethod
    def _check_bytes_data(cls, data: bytes, size: int):
        if not isinstance(data, bytes):
            raise IllegalFormatException("Data is not bytes")

        if len(data) != size:
            raise IllegalFormatException(f"Out of data: data_size({len(data)}) != size_to_read({size})")



# -*- coding: utf-8 -*-

# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

from iconcommons import Logger
from ...icon_constant import IISS_MAX_REWARD_RATE, IISS_ANNUAL_BLOCK, IISS_MONTH, IISS_LOG_TAG, PERCENTAGE_FOR_BETA_2

if TYPE_CHECKING:
    from ...iconscore.icon_score_context import IconScoreContext


class IssueFormula(object):

    def __init__(self, main_prep_count: int):
        self._handler: dict = {
            'prep': self._handle_icx_issue_formula_for_prep
        }
        # todo: in case of issuing from IISS_REV, get from the storage (not constant value)
        self.main_prep_count: int = main_prep_count

    def calculate(self, context: 'IconScoreContext', group: str, data: dict) -> int:
        handler = self._handler[group]
        # todo: as field name changed, this handler pattern not efficient. change the logic
        value = handler(context=context,
                        irep=data["irep"],
                        rrep=data["rrep"],
                        total_delegation=data["totalDelegation"])
        return value

    @staticmethod
    def calculate_rrep(rmin: int, rmax: int, rpoint: int, total_supply: int, total_delegated: int) -> int:
        stake_percentage: float = total_delegated / total_supply * IISS_MAX_REWARD_RATE
        if stake_percentage >= rpoint:
            return rmin

        first_operand: float = (rmax - rmin) / (rpoint ** 2)
        second_operand: float = (stake_percentage - rpoint) ** 2
        return int(first_operand * second_operand + rmin)

    @staticmethod
    def calculate_irep_per_block_contributor(irep: int) -> int:
        return int(irep * IISS_MONTH // (IISS_ANNUAL_BLOCK * 2))

    def _handle_icx_issue_formula_for_prep(self,
                                           context: 'IconScoreContext',
                                           irep: int,
                                           rrep: int,
                                           total_delegation: int) -> int:
        calculated_irep: int = self.calculate_irep_per_block_contributor(irep)
        beta_1: int = 0
        beta_2: int = 0
        if context.is_decentralized():
            beta_1: int = calculated_irep * self.main_prep_count
            beta_2: int = calculated_irep * PERCENTAGE_FOR_BETA_2 if context.term.total_delegated > 0 else 0

        temp_rrep = IssueFormula.calculate_temporary_reward_prep(rrep)
        beta_3: int = temp_rrep * total_delegation // (IISS_ANNUAL_BLOCK * IISS_MAX_REWARD_RATE)
        Logger.info("Calculated issue amount about this block. "
                    f"calculated_irep: {calculated_irep} irep: {irep} rrep: {temp_rrep} "
                    f"total_delegation: {total_delegation} "
                    f"beta1: {beta_1} beta2: {beta_2} beta3: {beta_3}", IISS_LOG_TAG)
        return beta_1 + beta_2 + beta_3

    @staticmethod
    def calculate_temporary_reward_prep(rrep: int) -> int:
        # todo: after eep and dapp is added, do not multiple 3 to beta3
        return rrep * 3



# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# link: https://github.com/Tierion/pymerkletools/

import hashlib
from typing import Union, Iterable, ByteString


class MerkleTree:
    hash_function = hashlib.sha3_256

    def __init__(self):
        self.leaves = list()
        self.levels = None
        self.is_ready = False
        self.reset_tree()

    def reset_tree(self):
        self.leaves = list()
        self.levels = None
        self.is_ready = False

    def add_leaf(self, values: Union[Iterable[ByteString], ByteString], do_hash=False):
        self.is_ready = False
        # check if single leaf
        if not isinstance(values, Iterable):
            values = [values]

        for v in values:
            if do_hash:
                v = self.hash_function(v).digest()
            v = bytes(v)
            self.leaves.append(v)

    def get_leaf(self, index):
        return self.leaves[index]

    def get_leaf_count(self):
        return len(self.leaves)

    def get_tree_ready_state(self):
        return self.is_ready

    def _calculate_next_level(self):
        solo_leave = None
        N = len(self.levels[0])  # number of leaves on the level
        if N % 2 == 1:  # if odd number of leaves on the level
            solo_leave = self.levels[0][-1]
            N -= 1

        new_level = []
        for l, r in zip(self.levels[0][0:N:2], self.levels[0][1:N:2]):
            new_level.append(self.hash_function(l+r).digest())
        if solo_leave is not None:
            new_level.append(solo_leave)
        self.levels = [new_level, ] + self.levels  # prepend new level

    def make_tree(self):
        self.is_ready = False
        if self.get_leaf_count() > 0:
            self.levels = [self.leaves, ]
            while len(self.levels[0]) > 1:
                self._calculate_next_level()
        self.is_ready = True

    def get_merkle_root(self):
        if self.is_ready:
            if self.levels is not None:
                return self.levels[0][0]
            else:
                return None
        else:
            return None

    def get_proof(self, index):
        if self.levels is None:
            return None
        elif not self.is_ready or index > len(self.leaves)-1 or index < 0:
            return None
        else:
            proof = []
            for x in range(len(self.levels) - 1, 0, -1):
                level_len = len(self.levels[x])
                if (index == level_len - 1) and (level_len % 2 == 1):  # skip if this is an odd end node
                    index = int(index / 2.)
                    continue
                is_right_node = index % 2
                sibling_index = index - 1 if is_right_node else index + 1
                sibling_pos = "left" if is_right_node else "right"
                sibling_value = self.levels[x][sibling_index]
                proof.append({sibling_pos: sibling_value})
                index = int(index / 2.)
            return proof

    @classmethod
    def validate_proof(cls, proof, target_hash, merkle_root):
        merkle_root = bytes(merkle_root)
        target_hash = bytes(target_hash)
        if len(proof) == 0:
            return target_hash == merkle_root
        else:
            proof_hash = target_hash
            for p in proof:
                try:
                    # the sibling is a left node
                    sibling = bytes(p['left'])
                    proof_hash = cls.hash_function(sibling + proof_hash).digest()
                except KeyError:
                    # the sibling is a right node
                    sibling = bytes(p['right'])
                    proof_hash = cls.hash_function(proof_hash + sibling).digest()
            return proof_hash == merkle_root



# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Optional, Iterable

from .data import Term
from .data.prep import PRep
from ..base.ComponentBase import StorageBase
from ..base.exception import InvalidParamsException
from ..utils.msgpack_for_db import MsgPackForDB

if TYPE_CHECKING:
    from ..base.address import Address
    from ..iconscore.icon_score_context import IconScoreContext
    from ..database.db import ContextDatabase


class Storage(StorageBase):
    PREFIX: bytes = b'prep'
    TERM_KEY: bytes = PREFIX + b'term'
    PREP_REGISTRATION_FEE_KEY: bytes = PREFIX + b'prf'

    def __init__(self, db: 'ContextDatabase'):
        super().__init__(db)
        self.prep_registration_fee: Optional[int] = None

    def open(self,
             context: 'IconScoreContext',
             prep_registration_fee: int):

        prep_reg_fee_from_db: Optional[int] = self.get_prep_registration_fee(context)
        if prep_reg_fee_from_db is None:
            self.put_prep_registration_fee(context, prep_registration_fee)
            self.prep_registration_fee = prep_registration_fee
        else:
            self.prep_registration_fee = prep_reg_fee_from_db

    def get_prep_registration_fee(self, context: 'IconScoreContext') -> Optional[int]:
        value: bytes = self._db.get(context, self.PREP_REGISTRATION_FEE_KEY)
        if value:
            data = MsgPackForDB.loads(value)
            version: int = data[0]
            assert version == 0
            prep_reg_fee: int = data[1]
            return prep_reg_fee
        else:
            return None

    def put_prep_registration_fee(self, context: 'IconScoreContext', prep_reg_fee: int):
        version = 0
        data: bytes = MsgPackForDB.dumps([version, prep_reg_fee])
        self._db.put(context, self.PREP_REGISTRATION_FEE_KEY, data)

    def get_prep(self, context: 'IconScoreContext', address: 'Address') -> 'PRep':
        key: bytes = PRep.make_key(address)
        value: bytes = self._db.get(context, key)

        if value is None:
            raise InvalidParamsException(f"P-Rep not found: {str(address)}")

        prep = PRep.from_bytes(value)
        assert address == prep.address

        return prep

    def put_prep(self, context: 'IconScoreContext', prep: 'PRep'):
        key: bytes = PRep.make_key(prep.address)
        value: bytes = prep.to_bytes(context.revision)
        self._db.put(context, key, value)

    def delete_prep(self, context: 'IconScoreContext', address: 'Address'):
        key: bytes = PRep.make_key(address)
        self._db.delete(context, key)

    def get_prep_iterator(self) -> Iterable['PRep']:
        with self._db.key_value_db.get_sub_db(PRep.PREFIX).iterator() as it:
            for key, value in it:
                if key[0] == 0x00 and len(key) == 21:
                    yield PRep.from_bytes(value)

    def put_term(self, context: 'IconScoreContext', term: 'Term'):
        value: bytes = MsgPackForDB.dumps(term.to_list())
        self._db.put(context, self.TERM_KEY, value)

    def get_term(self,
                 context: 'IconScoreContext') -> Optional['Term']:

        value: bytes = self._db.get(context, self.TERM_KEY)
        if value:
            total_elected_prep_delegated_snapshot: int = \
                context.storage.rc.get_total_elected_prep_delegated_snapshot()
            data: list = MsgPackForDB.loads(value)
            return Term.from_list(data,
                                  context.block.height,
                                  total_elected_prep_delegated_snapshot)



# -*- coding: utf-8 -*-
# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = 'RewardCalcProxy'

import asyncio
import concurrent.futures
import os
from subprocess import Popen
from typing import TYPE_CHECKING, Optional, Callable, Any, Tuple

from iconcommons.logger import Logger
from .message import *
from .message_queue import MessageQueue
from .server import IPCServer
from ....base.address import Address
from ....base.exception import TimeoutException
from ....icon_constant import RCStatus
from ....utils import bytes_to_hex

if TYPE_CHECKING:
    from .message import ReadyNotification, CalculateDoneNotification, NoneResponse

_TAG = "RCP"


class RewardCalcBlock(object):
    """Stores the latest commit status of reward calculator
    """

    def __init__(self, block_height: int, block_hash: bytes):
        self._block_height = block_height
        self._block_hash = block_hash

    @property
    def block_height(self) -> int:
        return self._block_height

    @property
    def block_hash(self) -> bytes:
        return self._block_hash


class RewardCalcProxy(object):
    """Communicates with Reward Calculator through UNIX Domain Socket

    """
    _DEFAULT_REWARD_CALCULATOR_PATH = "icon_rc"

    def __init__(self,
                 icon_rc_path: str,
                 ipc_timeout: int,
                 ready_callback: Callable[['ReadyNotification'], Any] = None,
                 calc_done_callback: Callable[['CalculateDoneNotification'], Any] = None):
        Logger.debug(tag=_TAG, msg="__init__() start")
        Logger.info(tag=_TAG, msg=f"ipc_timeout: {ipc_timeout}")

        self._loop = None
        self._ipc_server = IPCServer()
        self._message_queue: Optional['MessageQueue'] = None
        self._reward_calc: Optional[Popen] = None

        self._ready_future: Optional[asyncio.Future] = None

        self._ready_callback: Optional[Callable] = ready_callback
        self._calculate_done_callback: Optional[Callable] = calc_done_callback
        self._ipc_timeout = ipc_timeout
        self._icon_rc_path = icon_rc_path
        self._rc_block: Optional[RewardCalcBlock] = None

        Logger.debug(tag=_TAG, msg="__init__() end")

    def open(self, log_dir: str, sock_path: str, iiss_db_path: str, icon_rc_monitor: bool):
        Logger.debug(tag=_TAG, msg="open() start")

        self._loop = asyncio.get_event_loop()
        self._message_queue = MessageQueue(loop=self._loop,
                                           notify_message=(VersionResponse, CalculateResponse),
                                           notify_handler=self.notify_handler)
        self._ipc_server.open(self._loop, self._message_queue, sock_path)

        self.start_reward_calc(log_dir=log_dir, sock_path=sock_path, iiss_db_path=iiss_db_path,
                               icon_rc_monitor=icon_rc_monitor)
        self._ready_future = self._loop.create_future()

        Logger.debug(tag=_TAG, msg="open() end")

    def start(self):
        Logger.debug(tag=_TAG, msg="start() end")

        self._ipc_server.start()

        Logger.debug(tag=_TAG, msg="start() end")

    def stop(self):
        Logger.debug(tag=_TAG, msg="stop() start")

        self._stop_message_queue()
        self._ipc_server.stop()

        Logger.debug(tag=_TAG, msg="stop() end")

    def close(self):
        Logger.debug(tag=_TAG, msg="close() start")

        self._ipc_server.close()
        self.stop_reward_calc()

        self._message_queue = None
        self._loop = None
        self._rc_block = None

        Logger.debug(tag=_TAG, msg="close() end")

    def _stop_message_queue(self):
        Logger.info(tag=_TAG, msg="_stop_message_queue() start")

        request = NoneRequest()
        self._message_queue.put(request)

        Logger.info(tag=_TAG, msg="_stop_message_queue() end")

    def is_reward_calculator_ready(self) -> bool:
        return self._ready_future.done()

    def get_version(self):
        Logger.debug(tag=_TAG, msg="get_version() start")

        future: concurrent.futures.Future = \
            asyncio.run_coroutine_threadsafe(self._get_version(), self._loop)

        try:
            response: 'VersionResponse' = future.result(self._ipc_timeout)
        except asyncio.TimeoutError:
            future.cancel()
            raise TimeoutException("get_version message to RewardCalculator has timed-out")

        Logger.debug(tag=_TAG, msg=f"get_version() end: {response.version}")

        return response.version

    async def _get_version(self) -> 'VersionResponse':
        Logger.debug(tag=_TAG, msg="_get_version() start")

        request = VersionRequest()

        future: asyncio.Future = self._message_queue.put(request)
        await future

        Logger.debug(tag=_TAG, msg="_get_version() end")

        return future.result()

    def calculate(self, db_path: str, block_height: int) -> int:
        """Request RewardCalculator to calculate IScore for every account

        It is called on invoke thread

        :param db_path: the absolute path of iiss database
        :param block_height: The blockHeight when this request are sent to RewardCalculator
        """
        Logger.debug(tag=_TAG, msg="calculate() start")

        future: concurrent.futures.Future = \
            asyncio.run_coroutine_threadsafe(self._calculate(db_path, block_height), self._loop)

        try:
            response: 'CalculateResponse' = future.result(self._ipc_timeout)
        except asyncio.TimeoutError:
            future.cancel()
            raise TimeoutException("calculate message to RewardCalculator has timed-out")

        Logger.debug(tag=_TAG, msg=f"calculate() end: {response}")
        return response.status

    async def _calculate(self, db_path: str, block_height: int) -> 'CalculateResponse':
        Logger.debug(tag=_TAG, msg="_calculate() start")

        request = CalculateRequest(db_path, block_height)

        future: asyncio.Future = self._message_queue.put(request)
        await future

        Logger.debug(tag=_TAG, msg=f"_calculate() end")

        return future.result()

    def claim_iscore(self, address: 'Address',
                     block_height: int, block_hash: bytes,
                     tx_index: int, tx_hash: bytes) -> Tuple[int, int]:
        """Claim IScore of a given address

        It is called on invoke thread

        :param address: the address to claim
        :param block_height: the height of block which contains this claim tx
        :param block_hash: the hash of block which contains this claim tx
        :param tx_index: the index of claimIScore transaction which is contained in a block
        :param tx_hash: the hash of claimIScore transaction
        :return: [i-score(int), block_height(int)]
        :exception TimeoutException: The operation has timed-out
        """
        Logger.debug(
            tag=_TAG,
            msg=f"claim_iscore() start: "
                f"address({address}) block_height({block_height}) block_hash({block_hash.hex()})"
        )

        future: concurrent.futures.Future = asyncio.run_coroutine_threadsafe(
            self._claim_iscore(address, block_height, block_hash, tx_index, tx_hash), self._loop)

        try:
            response: 'ClaimResponse' = future.result(self._ipc_timeout)
        except asyncio.TimeoutError:
            future.cancel()
            raise TimeoutException("claim_iscore message to RewardCalculator has timed-out")

        Logger.debug(tag=_TAG, msg=f"claim_iscore() end: iscore({response.iscore})")

        return response.iscore, response.block_height

    async def _claim_iscore(self, address: 'Address',
                            block_height: int, block_hash: bytes,
                            tx_index: int, tx_hash: bytes) -> 'ClaimResponse':
        Logger.debug(
            tag=_TAG,
            msg=f"_claim_iscore() start: "
                f"address={address} "
                f"block_height={block_height} "
                f"block_hash={bytes_to_hex(block_hash)} "
                f"tx_index={tx_index} "
                f"tx_hash={bytes_to_hex(tx_hash)}"
        )
        request = ClaimRequest(address, block_height, block_hash, tx_index, tx_hash)

        future: asyncio.Future = self._message_queue.put(request)
        await future

        Logger.debug(tag=_TAG, msg=f"_claim_iscore() end")

        return future.result()

    def commit_claim(self, success: bool, address: 'Address',
                     block_height: int, block_hash: bytes,
                     tx_index: int, tx_hash: bytes):
        Logger.debug(
            tag=_TAG,
            msg=f"commit_claim() start: "
                f"success={success} "
                f"address={address} "
                f"block_height={block_height} "
                f"block_hash={bytes_to_hex(block_hash)} "
                f"tx_index={tx_index} "
                f"tx_hash={bytes_to_hex(tx_hash)}"
        )

        future: concurrent.futures.Future = asyncio.run_coroutine_threadsafe(
            self._commit_claim(success, address, block_height, block_hash, tx_index, tx_hash),
            self._loop
        )

        try:
            future.result(self._ipc_timeout)

        except asyncio.TimeoutError:
            future.cancel()
            raise TimeoutException("COMMIT_CLAIM message to RewardCalculator has timed-out")

        Logger.debug(tag=_TAG, msg="commit_claim() end")

    async def _commit_claim(self, success: bool, address: 'Address',
                            block_height: int, block_hash: bytes,
                            tx_index: int, tx_hash: bytes) -> 'CommitClaimResponse':
        Logger.debug(
            tag=_TAG,
            msg=f"_commit_claim() start: "
                f"success={success} "
                f"address={address} "
                f"block_height={block_height} "
                f"block_hash={bytes_to_hex(block_hash)} "
                f"tx_index={tx_index} "
                f"tx_hash={bytes_to_hex(tx_hash)}"
        )

        request = CommitClaimRequest(success, address, block_height, block_hash, tx_index, tx_hash)

        future: asyncio.Future = self._message_queue.put(request)
        await future

        Logger.debug(tag=_TAG, msg="_commit_claim() end")

        return future.result()

    def query_iscore(self, address: 'Address') -> Tuple[int, int]:
        """Returns the I-Score of a given address

        It should be called on query thread

        :param address: the address to query
        :return: [i-score(int), block_height(int)]
        :exception TimeoutException: The operation has timed-out
        """
        assert isinstance(address, Address)

        Logger.debug(tag=_TAG, msg="query_iscore() start")

        future: concurrent.futures.Future = asyncio.run_coroutine_threadsafe(
            self._query_iscore(address), self._loop)

        try:
            response: 'QueryResponse' = future.result(self._ipc_timeout)
        except asyncio.TimeoutError:
            future.cancel()
            raise TimeoutException("query_iscore message to RewardCalculator has timed-out")

        Logger.debug(tag=_TAG, msg="query_iscore() end")

        return response.iscore, response.block_height

    async def _query_iscore(self, address: 'Address') -> 'QueryResponse':
        """

        :param address:
        :return: [iscore(int), block_height(int)]
        """
        Logger.debug(tag=_TAG, msg="_query_iscore() start")

        request = QueryRequest(address)

        future: asyncio.Future = self._message_queue.put(request)
        await future

        Logger.debug(tag=_TAG, msg="_query_iscore() end")

        return future.result()

    def query_calculate_status(self) -> tuple:
        Logger.debug(tag=_TAG, msg="query_calculate_status() start")

        future: concurrent.futures.Future = asyncio.run_coroutine_threadsafe(
            self._query_calculate_status(), self._loop)

        try:
            response: QueryCalculateStatusResponse = future.result(self._ipc_timeout)
        except asyncio.TimeoutError:
            future.cancel()
            raise TimeoutException("query_calculate_status message to RewardCalculator has timed-out")

        Logger.debug(tag=_TAG, msg="query_calculate_status() end")

        return response.status, response.block_height

    async def _query_calculate_status(self) -> 'QueryCalculateStatusResponse':
        Logger.debug(tag=_TAG, msg="_query_calculate_status() start")

        request = QueryCalculateStatusRequest()

        future: asyncio.Future = self._message_queue.put(request)
        await future

        Logger.debug(tag=_TAG, msg="_query_calculate_status() end")

        return future.result()

    def query_calculate_result(self, block_height) -> tuple:
        Logger.debug(tag=_TAG, msg="query_calculate_result() start")

        future: concurrent.futures.Future = asyncio.run_coroutine_threadsafe(
            self._query_calculate_result(block_height), self._loop)

        try:
            response: QueryCalculateResultResponse = future.result(self._ipc_timeout)
        except asyncio.TimeoutError:
            future.cancel()
            raise TimeoutException("query_calculate_result message to RewardCalculator has timed-out")

        Logger.debug(tag=_TAG, msg="query_calculate_result() end")

        return response.status, response.block_height, response.iscore, response.state_hash

    async def _query_calculate_result(self, block_height) -> 'QueryCalculateResultResponse':
        Logger.debug(tag=_TAG, msg="_query_calculate_result() start")

        request = QueryCalculateResultRequest(block_height)

        future: asyncio.Future = self._message_queue.put(request)
        await future

        Logger.debug(tag=_TAG, msg="_query_calculate_result() end")

        return future.result()

    def commit_block(self, success: bool, block_height: int, block_hash: bytes) -> tuple:
        """Notify reward calculator of block confirmation

        It is called on invoke thread

        :param success: true for success, false for failure
        :param block_height: the height of block
        :param block_hash: the hash of block
        :return: [success(bool), block_height(int), block_hash(bytes)]
        :exception TimeoutException: The operation has timed-out
        """
        Logger.debug(
            tag=_TAG,
            msg=f"commit_block() start: success={success}, "
                f"block_height={block_height}, "
                f"block_hash={bytes_to_hex(block_hash)}"
        )

        future: concurrent.futures.Future = asyncio.run_coroutine_threadsafe(
            self._commit_block(success, block_height, block_hash), self._loop)

        try:
            response: 'CommitBlockResponse' = future.result(self._ipc_timeout)
        except asyncio.TimeoutError:
            future.cancel()
            raise TimeoutException("commit_block message to RewardCalculator has timed-out")

        Logger.debug(tag=_TAG, msg=f"commit_block() end. response: {response}")

        return response.success, response.block_height, response.block_hash

    async def _commit_block(self, success: bool, block_height: int, block_hash: bytes) -> 'CommitBlockResponse':
        # Logger.debug(tag=_TAG, msg="_commit_block() start")

        request = CommitBlockRequest(success, block_height, block_hash)

        future: asyncio.Future = self._message_queue.put(request)
        await future

        # Logger.debug(tag=_TAG, msg="_commit_block() end")

        return future.result()

    def init_reward_calculator(self, block_height: int) -> int:
        Logger.debug(tag=_TAG, msg=f"init_reward_calculator() start: block_height={block_height}")

        future: concurrent.futures.Future = asyncio.run_coroutine_threadsafe(
            self._init_reward_calculator(block_height), self._loop)

        try:
            response: InitResponse = future.result(self._ipc_timeout)
        except asyncio.TimeoutError:
            future.cancel()
            raise TimeoutException("query_calculate_result message to RewardCalculator has timed-out")

        Logger.debug(tag=_TAG, msg="query_calculate_result() end")

        return response.success

    async def _init_reward_calculator(self, block_height: int):
        Logger.debug(tag=_TAG, msg=f"init_reward_calculator() start: block_height={block_height}")

        request = InitRequest(block_height)

        future: asyncio.Future = self._message_queue.put(request)
        await future

        Logger.debug(tag=_TAG, msg="init_reward_calculator() end")

        return future.result()

    def rollback(self, block_height: int, block_hash: bytes) -> Tuple[bool, int, bytes]:
        """Request reward calculator to rollback the DB of the reward calculator to the specific block height.

        Reward calculator DOES NOT process other messages while processing ROLLBACK message

        :param block_height:
        :param block_hash:
        :return:
        """

        Logger.debug(
            tag=_TAG,
            msg=f"rollback() start: block_height={block_height}, block_hash={bytes_to_hex(block_hash)}"
        )

        future: concurrent.futures.Future = asyncio.run_coroutine_threadsafe(
            self._rollback(block_height, block_hash), self._loop)

        try:
            response: 'RollbackResponse' = future.result(self._ipc_timeout)
        except asyncio.TimeoutError:
            future.cancel()
            raise TimeoutException("rollback message to RewardCalculator has timed-out")

        Logger.debug(tag=_TAG, msg=f"rollback() end. response: {response}")

        return response.success, response.block_height, response.block_hash

    async def _rollback(self, block_height: int, block_hash: bytes) -> 'RollbackResponse':
        Logger.debug(tag=_TAG, msg="_rollback() start")

        request = RollbackRequest(block_height, block_hash)

        future: asyncio.Future = self._message_queue.put(request)
        await future

        Logger.debug(tag=_TAG, msg="_rollback() end")

        return future.result()

    def ready_handler(self, response: 'ReadyNotification'):
        Logger.debug(tag=_TAG, msg=f"ready_handler() start {response}")

        if self._ready_callback is not None:
            self._ready_callback(response)

        self._ready_future.set_result(RCStatus.READY)
        self._rc_block = RewardCalcBlock(response.block_height, response.block_height)

    def get_ready_future(self):
        return self._ready_future

    def calculate_done_handler(self, response: 'Response'):
        Logger.debug(tag=_TAG, msg=f"calculate_done_handler() start {response}")
        if self._calculate_done_callback is not None:
            self._calculate_done_callback(response)

    def notify_handler(self, response: 'Response'):
        Logger.debug(tag=_TAG, msg=f"notify_handler() start {type(response)}")
        if isinstance(response, ReadyNotification):
            self.ready_handler(response=response)
        elif isinstance(response, CalculateDoneNotification):
            self.calculate_done_handler(response=response)

    def start_reward_calc(self, log_dir: str, sock_path: str, iiss_db_path: str, icon_rc_monitor: bool):
        """ Start reward calculator process

        :param log_dir: log directory
        :param sock_path: unix domain socket path for IPC
        :param iiss_db_path: IISS data DB path
        :param icon_rc_monitor: Boolean which determines Opening RC monitor channel (default True)
        :return: void
        """
        Logger.debug(tag=_TAG, msg=f'run reward calc')

        iscore_db_path, _ = os.path.split(iiss_db_path)
        iscore_db_path = os.path.join(iscore_db_path, 'rc')
        log_path = os.path.join(log_dir, 'rc.log')
        reward_calculator_path: str = self._get_reward_calculator_path(self._icon_rc_path)

        if self._reward_calc is None:
            args = [
                reward_calculator_path,
                "-client",
                "-db-count", "16",
                "-db", f"{iscore_db_path}",
                "-iissdata", f"{iiss_db_path}",
                "-ipc-addr", f"{sock_path}",
                "-log-file", f"{log_path}",
            ]
            if icon_rc_monitor is True:
                args.append("-monitor")

            Logger.info(tag=_TAG, msg=f"cmd={' '.join(args)}")
            self._reward_calc = Popen(args)

        Logger.debug(tag=_TAG, msg="start_reward_calc() end")

    def _get_reward_calculator_path(self, path: str) -> str:
        command = self._DEFAULT_REWARD_CALCULATOR_PATH

        if isinstance(path, str):
            if os.path.isdir(path):
                return os.path.join(path, command)
            elif os.path.isfile(path):
                return path

        return command

    def stop_reward_calc(self):
        """ Stop reward calculator process

        :return: void
        """
        Logger.debug(tag=_TAG, msg='stop reward calc')

        if self._reward_calc is not None:
            self._reward_calc.kill()
            self._reward_calc = None

    def get_commit_block(self) -> Optional[Tuple[int, bytes]]:
        if self._rc_block is None:
            return None

        return self._rc_block.block_height, self._rc_block.block_hash



# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import warnings
from struct import pack, unpack
from typing import TYPE_CHECKING, Optional, Tuple

from ..base.ComponentBase import StorageBase
from ..base.address import Address, ICON_EOA_ADDRESS_BYTES_SIZE, ICON_CONTRACT_ADDRESS_BYTES_SIZE
from ..base.exception import InvalidParamsException, AccessDeniedException
from ..icon_constant import DEFAULT_BYTE_SIZE, Revision, ZERO_TX_HASH, DeployState, DeployType

if TYPE_CHECKING:
    from ..iconscore.icon_score_context import IconScoreContext


class IconScoreDeployTXParams(object):
    _VERSION = 0
    _STRUCT_FMT = f'>BBI{ICON_CONTRACT_ADDRESS_BYTES_SIZE}s{DEFAULT_BYTE_SIZE}s'
    _PIVOT_SIZE = 1 + 1 + 4 + ICON_CONTRACT_ADDRESS_BYTES_SIZE + DEFAULT_BYTE_SIZE

    # leveldb IconScoreDeployTXParams value structure
    # (bigendian, 1 + 1 + 4+ ICON_CONTRACT_ADDRESS_BYTES_SIZE + DEFAULT_BYTE_SIZE + data_bytes)
    # version(1)
    # | deploy_type(1)
    # | deploy_data_length(4)
    # | score_address(21)
    # | tx_hash(DEFAULT_BYTE_SIZE)
    # | deploy_data_value(deploy_data_length)

    def __init__(self,
                 tx_hash: bytes,
                 deploy_type: 'DeployType',
                 score_address: 'Address',
                 deploy_data: dict):
        # key
        self._tx_hash = tx_hash
        # value
        self._score_address = score_address
        self._deploy_type = deploy_type
        self._deploy_data = deploy_data

    @property
    def tx_hash(self) -> bytes:
        return self._tx_hash

    @property
    def deploy_type(self) -> 'DeployType':
        return self._deploy_type

    @property
    def score_address(self) -> 'Address':
        return self._score_address

    @property
    def deploy_data(self) -> dict:
        return self._deploy_data

    @staticmethod
    def from_bytes(buf: bytes) -> 'IconScoreDeployTXParams':
        """Create IconScoreDeployTXParams object from bytes data

        :param buf: (bytes) bytes data including IconScoreDeployTXParams information
        :return: (IconScoreDeployTXParams) IconScoreDeployTXParams object
        """

        version, deploy_type, deploy_data_length, score_addr_bytes, hash_bytes = unpack(
            IconScoreDeployTXParams._STRUCT_FMT, buf[:IconScoreDeployTXParams._PIVOT_SIZE])

        json_str_deploy_data_bytes = unpack(
            f'{deploy_data_length}s', buf[IconScoreDeployTXParams._PIVOT_SIZE:])[0]

        score_address = Address.from_bytes(score_addr_bytes)
        deploy_data = json.loads(json_str_deploy_data_bytes.decode())

        tx_params = IconScoreDeployTXParams(hash_bytes, DeployType(deploy_type), score_address, deploy_data)
        return tx_params

    def to_bytes(self) -> bytes:
        """Convert block object to bytes

        :return: data including information of IconScoreDeployInfo object
        """

        # for extendability
        json_str_deploy_data: str = json.dumps(self._deploy_data)
        deploy_data: bytes = json_str_deploy_data.encode(encoding='utf-8')
        deploy_data_length: int = len(deploy_data)

        bytes_var1 = pack(
            IconScoreDeployTXParams._STRUCT_FMT,
            self._VERSION, self._deploy_type.value, deploy_data_length,
            self._score_address.to_bytes(), self._tx_hash)
        bytes_var2 = pack(f'>{deploy_data_length}s', deploy_data)

        return bytes_var1 + bytes_var2


class IconScoreDeployInfo(object):
    """
    leveldb IconScoreDeployInfo value structure
    (bigendian, 1 + 1 + ICON_CONTRACT_ADDRESS_BYTES_SIZE + ICON_EOA_ADDRESS_BYTES_SIZE +
    DEFAULT_BYTE_SIZE + DEFAULT_BYTE_SIZE bytes)
    | version(1)
    | deploystate(1)
    | score_address(ICON_CONTRACT_ADDRESS_BYTES_SIZE)
    | owner(ICON_EOA_ADDRESS_BYTES_SIZE)
    | current_tx_hash(DEFAULT_BYTE_SIZE)
    | next_tx_hash(DEFAULT_BYTE_SIZE)
    """

    _VERSION = 0
    _STRUCT_FMT = f'>BB' \
        f'{ICON_CONTRACT_ADDRESS_BYTES_SIZE}s' \
        f'{ICON_EOA_ADDRESS_BYTES_SIZE}s' \
        f'{DEFAULT_BYTE_SIZE}s' \
        f'{DEFAULT_BYTE_SIZE}s'

    def __init__(self,
                 score_address: 'Address',
                 deploy_state: 'DeployState',
                 owner: 'Address',
                 current_tx_hash: bytes,
                 next_tx_hash: bytes):
        assert isinstance(current_tx_hash, bytes) and len(current_tx_hash) == 32
        assert isinstance(next_tx_hash, bytes) and len(next_tx_hash) == 32

        # key
        self._score_address = score_address

        # value
        self.deploy_state = deploy_state
        self.owner = owner
        self.current_tx_hash = current_tx_hash
        self.next_tx_hash = next_tx_hash

    def __str__(self):
        return f"score_address={self.score_address}, " \
               f"owner={self.owner}, " \
               f"state={self.deploy_state}, " \
               f"current_tx_hash={self.current_tx_hash}, " \
               f"next_tx_hash={self.next_tx_hash}"

    @property
    def score_address(self):
        return self._score_address

    @staticmethod
    def from_bytes(buf: bytes) -> 'IconScoreDeployInfo':
        """Create IconScoreDeployInfo object from bytes data

        :param buf: (bytes) bytes data including IconScoreDeployInfo information
        :return: (IconScoreDeployInfo) IconScoreDeployInfo object
        """

        bytes_params = unpack(IconScoreDeployInfo._STRUCT_FMT, buf)
        # version: int = bytes_params[0]
        deploy_state: int = bytes_params[1]
        score_address: 'Address' = Address.from_bytes(bytes_params[2])
        owner_address: 'Address' = Address.from_bytes(bytes_params[3])
        current_tx_hash: bytes = bytes_params[4]
        next_tx_hash: bytes = bytes_params[5]

        # DeployState 2 is deprecated, this code is needed for backward compatibility.
        if deploy_state == 2:
            deploy_state = DeployState.ACTIVE

        return IconScoreDeployInfo(
            score_address, DeployState(deploy_state), owner_address, current_tx_hash, next_tx_hash)

    def to_bytes(self) -> bytes:
        """Convert IconScoreDeployInfo object to bytes

        :return: data including information of IconScoreDeployInfo object
        """
        return pack(
            self._STRUCT_FMT, self._VERSION, self.deploy_state.value,
            self._score_address.to_bytes(), self.owner.to_bytes(),
            self.current_tx_hash, self.next_tx_hash)


class Storage(StorageBase):
    """Store deploy_int and tx_params on LevelDB.
    """

    _DEPLOY_STORAGE_PREFIX = b'isds|'
    _DEPLOY_STORAGE_DEPLOY_INFO_PREFIX = _DEPLOY_STORAGE_PREFIX + b'di|'
    _DEPLOY_STORAGE_DEPLOY_TX_PARAMS_PREFIX = _DEPLOY_STORAGE_PREFIX + b'dtp|'

    def put_deploy_info_and_tx_params(self,
                                      context: 'IconScoreContext',
                                      score_address: 'Address',
                                      deploy_type: 'DeployType',
                                      owner: 'Address',
                                      tx_hash: bytes,
                                      deploy_data: dict) -> None:
        prev_tx_params = self.get_deploy_tx_params(context, tx_hash)
        if prev_tx_params is not None:
            raise InvalidParamsException(f'deploy_params already exists: {tx_hash}')

        # Save DeployTXParams to stateDB
        tx_params = IconScoreDeployTXParams(tx_hash, deploy_type, score_address, deploy_data)
        self.put_deploy_tx_params(context, tx_params)

        deploy_info = self.get_deploy_info(context, score_address)
        if deploy_info is None:
            # SCORE install case
            deploy_info = IconScoreDeployInfo(
                score_address, DeployState.INACTIVE, owner, ZERO_TX_HASH, tx_hash)
        else:
            # SCORE update case
            if deploy_info.owner != owner:
                raise AccessDeniedException(f'Invalid owner: {deploy_info.owner} != {owner}')

            # If the previous DeployTXParams has exists, remove it before deploying
            if deploy_info.next_tx_hash != ZERO_TX_HASH:
                self.delete_deploy_tx_params(context, deploy_info.next_tx_hash)

            deploy_info.next_tx_hash = tx_hash

        # Save DeployInfo to stateDB
        self.put_deploy_info(context, deploy_info)

    def update_score_info(self,
                          context: 'IconScoreContext',
                          score_address: 'Address',
                          tx_hash: bytes) -> None:

        deploy_info = self.get_deploy_info(context, score_address)
        if deploy_info is None:
            raise InvalidParamsException(f'deploy_info is None: {score_address}')

        next_tx_hash = deploy_info.next_tx_hash
        # have to match next_tx_hash and tx_hash
        # tx_hash is None -> builtin install
        if tx_hash is not None and tx_hash != next_tx_hash:
            raise InvalidParamsException(
                f'Invalid update: tx_hash({tx_hash}) != next_tx_hash({next_tx_hash})')

        deploy_info.current_tx_hash = next_tx_hash
        deploy_info.next_tx_hash = ZERO_TX_HASH
        deploy_info.deploy_state = DeployState.ACTIVE
        self.put_deploy_info(context, deploy_info)

        tx_params = self.get_deploy_tx_params(context, deploy_info.current_tx_hash)
        if tx_params is None:
            raise InvalidParamsException(f'tx_params is None: {deploy_info.current_tx_hash}')

    def put_deploy_info(self, context: Optional['IconScoreContext'], deploy_info: 'IconScoreDeployInfo') -> None:
        """

        :param context:
        :param deploy_info:
        :return:
        """
        key: bytes = self._create_db_key(
            self._DEPLOY_STORAGE_DEPLOY_INFO_PREFIX, deploy_info.score_address.to_bytes())
        value: bytes = deploy_info.to_bytes()

        self._db.put(context, key, value)

    def get_deploy_info(self, context: Optional['IconScoreContext'], score_address: 'Address') \
            -> Optional['IconScoreDeployInfo']:

        key: bytes = self._create_db_key(
            self._DEPLOY_STORAGE_DEPLOY_INFO_PREFIX, score_address.to_bytes())
        data: bytes = self._db.get(context, key)
        if data is None:
            return None

        return IconScoreDeployInfo.from_bytes(data)

    def put_deploy_tx_params(self, context: 'IconScoreContext', deploy_tx_params: 'IconScoreDeployTXParams') -> None:
        """

        :param context:
        :param deploy_tx_params:
        :return:
        """
        key: bytes = self._create_db_key(self._DEPLOY_STORAGE_DEPLOY_TX_PARAMS_PREFIX, deploy_tx_params.tx_hash)
        value = deploy_tx_params.to_bytes()
        self._db.put(context, key, value)

    def get_deploy_tx_params(self, context: 'IconScoreContext', tx_hash: bytes) -> Optional['IconScoreDeployTXParams']:
        key: bytes = self._create_db_key(self._DEPLOY_STORAGE_DEPLOY_TX_PARAMS_PREFIX, tx_hash)
        value: bytes = self._db.get(context, key)

        if value is None:
            return None

        return IconScoreDeployTXParams.from_bytes(value)

    def delete_deploy_tx_params(self, context: 'IconScoreContext', tx_hash: bytes):
        """Delete DeployTXParams indicated by tx_hash from stateDB.

        :param context:
        :param tx_hash:
        :return:
        """
        if context.revision >= Revision.TWO.value:
            key: bytes = self._create_db_key(
                self._DEPLOY_STORAGE_DEPLOY_TX_PARAMS_PREFIX, tx_hash)
        else:
            key: bytes = tx_hash

        self._db.delete(context, key)

    @staticmethod
    def _create_db_key(prefix: bytes, src_key: bytes) -> bytes:
        return prefix + src_key

    def get_tx_hashes_by_score_address(self,
                                       context: 'IconScoreContext',
                                       score_address: 'Address') -> Tuple[Optional[bytes], Optional[bytes]]:
        warnings.warn("legacy function don't use.", DeprecationWarning, stacklevel=2)
        deploy_info = self.get_deploy_info(context, score_address)
        if deploy_info:
            current_tx_hash = deploy_info.current_tx_hash
            if current_tx_hash == ZERO_TX_HASH:
                # Keep compatibility for old blocks
                current_tx_hash = None

            next_tx_hash = deploy_info.next_tx_hash
            if next_tx_hash == ZERO_TX_HASH:
                # Keep compatibility for old blocks
                next_tx_hash = None

            return current_tx_hash, next_tx_hash

        return None, None

    def get_score_address_by_tx_hash(self,
                                     context: 'IconScoreContext',
                                     tx_hash: bytes) -> Optional['Address']:
        warnings.warn("legacy function don't use.", DeprecationWarning, stacklevel=2)
        tx_params = self.get_deploy_tx_params(context, tx_hash)
        if tx_params:
            return tx_params.score_address

        return None



# -*- coding: utf-8 -*-
# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from typing import Callable, Any, Optional, Dict

from iconcommons.logger import Logger
from iconservice.base.exception import InvalidParamsException, ServiceNotReadyException
from iconservice.icon_constant import RCStatus
from .message import Request, Response, MessageType


class MessageQueue(object):
    def __init__(self, loop, notify_message: tuple = None, notify_handler: Callable[['Response'], Any] = None):
        if notify_handler is None and notify_message is not None:
            raise InvalidParamsException("Failed to construct MessageQueue instance."
                                         "If notify_message is not None, notify_handler is mandatory parameter")
        self._loop = loop
        self._requests = asyncio.Queue()
        self._msg_id_to_future: Dict[int, asyncio.Future] = {}
        self.notify_message: Optional[tuple] = notify_message
        self.notify_handler = notify_handler
        self._rc_status = RCStatus.NOT_READY

    async def get(self) -> 'Request':
        return await self._requests.get()

    def put(self, request, wait_for_response: bool = True) -> Optional[asyncio.Future]:
        assert isinstance(request, Request)

        self._requests.put_nowait(request)

        if wait_for_response:
            future: asyncio.Future = self._loop.create_future()
            self._msg_id_to_future[request.msg_id] = future
            return future

    def task_done(self):
        self._requests.task_done()

    def message_handler(self, response: 'Response'):
        msg_type: MessageType = getattr(response, "MSG_TYPE")

        if self._rc_status == RCStatus.NOT_READY:
            if msg_type == MessageType.READY:
                self._rc_status = RCStatus.READY
            else:
                raise ServiceNotReadyException(f"Ready notification did not arrive: {response}")

        if response.is_notification():
            self.notify_handler(response)
            return

        try:
            self.put_response(response)
        except KeyError:
            Logger.warning(f"Unexpected response arrived.  Respond Id : {response.msg_id}")

    def put_response(self, response: 'Response'):
        msg_id: int = response.msg_id

        future: asyncio.Future = self._msg_id_to_future[msg_id]

        del self._msg_id_to_future[msg_id]

        future.set_result(response)



# -*- coding: utf-8 -*-
# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ("Term", "PRepSnapshot")

import copy
from typing import TYPE_CHECKING, List, Iterable, Optional, Dict

from iconcommons.logger import Logger
from ... import utils
from ...base.exception import AccessDeniedException
from ...icon_constant import PRepStatus, PenaltyReason, TermFlag
from ...utils import bytes_to_hex
from ...utils.hashing.hash_generator import RootHashGenerator

if TYPE_CHECKING:
    from ...base.address import Address
    from .prep import PRep


class PRepSnapshot(object):
    """Contains P-Rep address and the delegated amount when this term started
    """

    def __init__(self, address: 'Address', delegated: int):
        self._address = address
        self._delegated = delegated

    @property
    def address(self) -> 'Address':
        return self._address

    @property
    def delegated(self) -> int:
        return self._delegated

    def __eq__(self, other: 'PRepSnapshot') -> bool:
        return self.address == other.address \
               and self.delegated == other.delegated


class Term(object):
    """Manages P-Rep Term information
    """

    TAG = "TERM"
    _VERSION = 0

    def __init__(self,
                 sequence: int,
                 start_block_height: int,
                 period: int,
                 irep: int,
                 total_supply: int,
                 total_delegated: int):
        self._sequence = sequence

        self._start_block_height = start_block_height
        self._period = period

        self._irep = irep
        self._total_supply = total_supply
        self._total_delegated = total_delegated
        self._total_elected_prep_delegated = 0
        self._total_elected_prep_delegated_snapshot: int = 0

        self._main_preps: List['PRepSnapshot'] = []
        self._sub_preps: List['PRepSnapshot'] = []
        self._preps_dict: Dict['Address', 'PRepSnapshot'] = {}

        # made from main P-Rep addresses
        self._merkle_root_hash: Optional[bytes] = None
        self._is_frozen: bool = False
        self._flags: 'TermFlag' = TermFlag.NONE

    @property
    def flags(self) -> 'TermFlag':
        return self._flags

    def is_dirty(self):
        return utils.is_any_flag_on(self._flags, TermFlag.ALL)

    def on_main_prep_changed(self, flag: 'TermFlag'):
        self._check_access_permission()
        self._flags |= flag

    def is_frozen(self) -> bool:
        return self._is_frozen

    def freeze(self):
        self._is_frozen = True
        self._flags = TermFlag.NONE

    def _check_access_permission(self):
        if self.is_frozen():
            raise AccessDeniedException("Term access denied")

    def __bool__(self) -> bool:
        return True

    def __str__(self) -> str:
        return \
            f"Term:" \
            f"seq={self._sequence} " \
            f"start_block_height={self._start_block_height} " \
            f"period={self._period} " \
            f"irep={self._irep}" \
            f"total_supply={self._total_supply} " \
            f"total_delegated={self._total_delegated} " \
            f"total_elected_prep_delegated={self._total_elected_prep_delegated} " \
            f"_total_elected_prep_delegated_snapshot={self._total_elected_prep_delegated_snapshot} " \
            f"root_hash={bytes_to_hex(self._merkle_root_hash)}"

    def __contains__(self, address: 'Address') -> bool:
        """Check whether the given address is an elected P-Rep

        :param address: P-Rep address
        :return: True(elected P-Rep), False(non-elected P-Rep)
        """
        return address in self._preps_dict

    def __len__(self) -> int:
        """Return the number of main P-Reps and sub P-Reps

        :return:
        """
        return len(self._preps_dict)

    def __eq__(self, other: 'Term') -> bool:
        return isinstance(other, Term) \
            and self._sequence == other._sequence \
            and self._start_block_height == other._start_block_height \
            and self._period == other._period \
            and self._irep == other._irep \
            and self._total_supply == other._total_supply \
            and self._total_delegated == other._total_delegated \
            and self._total_elected_prep_delegated == other._total_elected_prep_delegated \
            and self._total_elected_prep_delegated_snapshot == other._total_elected_prep_delegated_snapshot \
            and self._main_preps == other._main_preps \
            and self._sub_preps == other._sub_preps \
            and self._preps_dict == other._preps_dict \
            and self._merkle_root_hash == other._merkle_root_hash

    def is_main_prep(self, address: 'Address') -> bool:
        for prep_snapshot in self._main_preps:
            if address == prep_snapshot.address:
                return True

        return False

    @property
    def sequence(self) -> int:
        return self._sequence

    @property
    def start_block_height(self) -> int:
        return self._start_block_height

    @property
    def end_block_height(self) -> int:
        return self._start_block_height + self._period - 1

    @property
    def period(self) -> int:
        return self._period

    @property
    def irep(self) -> int:
        """Returns weighted average irep used during a term

        :return: weighted average irep that is calculated with ireps submitted by 22 Main P-Reps
        """
        return self._irep

    @property
    def total_supply(self) -> int:
        return self._total_supply

    # @total_supply.setter
    # def total_supply(self, value: int):
    #     self._total_supply = value

    @property
    def total_delegated(self) -> int:
        """Total amount of delegation which all active P-Reps got when this term is started
        """
        return self._total_delegated

    @property
    def total_elected_prep_delegated(self) -> int:
        """The sum of delegated amount which only main and sub P-Reps got

        :return:
        """
        return self._total_elected_prep_delegated

    @property
    def total_elected_prep_delegated_snapshot(self) -> int:
        """The sum of delegated amount which only main and sub P-Reps got on start Term

        :return:
        """
        return self._total_elected_prep_delegated_snapshot

    @property
    def main_preps(self) -> List['PRepSnapshot']:
        return self._main_preps

    @property
    def sub_preps(self) -> List['PRepSnapshot']:
        return self._sub_preps

    @property
    def preps(self) -> Iterable['PRepSnapshot']:
        for snapshot in self._main_preps:
            yield snapshot
        for snapshot in self._sub_preps:
            yield snapshot

    @property
    def root_hash(self) -> bytes:
        assert isinstance(self._merkle_root_hash, bytes)
        return self._merkle_root_hash

    def is_in_term(self, block_height: int):
        return self.start_block_height <= block_height <= self.end_block_height

    def set_preps(self,
                  it: Iterable['PRep'],
                  main_prep_count: int,
                  elected_prep_count: int):
        """Set elected P-Rep data to term

        This method MUST BE called at the end of the term

        :param it:
        :param main_prep_count:
        :param elected_prep_count:
        :return:
        """

        self._main_preps.clear()
        self._sub_preps.clear()
        self._preps_dict.clear()

        # Main and sub P-Reps
        total_elected_prep_delegated: int = 0

        for i, prep in enumerate(it):
            if i >= elected_prep_count:
                break

            # active and no penalty
            assert prep.is_electable()

            snapshot = PRepSnapshot(prep.address, prep.delegated)
            total_elected_prep_delegated += prep.delegated

            if len(self._main_preps) < main_prep_count:
                self._main_preps.append(snapshot)
            else:
                self._sub_preps.append(snapshot)

            self._preps_dict[snapshot.address] = snapshot

        self._total_elected_prep_delegated_snapshot = total_elected_prep_delegated
        self._total_elected_prep_delegated = total_elected_prep_delegated

        self._generate_root_hash()
        self._flags = TermFlag.NONE

    def update_invalid_elected_preps(self, invalid_elected_preps: Iterable['PRep']):
        """Update main and sub P-Reps with invalid elected P-Reps

        :param invalid_elected_preps:
            elected P-Reps that cannot keep governance during this term as their penalties
        :return:
        """
        self._check_access_permission()
        flags: 'TermFlag' = TermFlag.NONE

        for prep in invalid_elected_preps:
            if self._remove_invalid_main_prep(prep) >= 0:
                flags |= TermFlag.MAIN_PREPS | TermFlag.SUB_PREPS
                continue
            if self._remove_invalid_sub_prep(prep) >= 0:
                flags |= TermFlag.SUB_PREPS
                continue

            raise AssertionError(f"{prep.address} not in elected P-Reps: {self}")

        if utils.is_all_flag_on(flags, TermFlag.MAIN_PREPS):
            self._generate_root_hash()

        self._flags |= flags

    def _remove_invalid_main_prep(self, invalid_prep: 'PRep') -> int:
        """Replace an invalid main P-Rep with the top-ordered sub P-Rep

        :param invalid_prep: an invalid main P-Rep
        :return: the index of an given address in self._main_preps
            if not exists, returns -1
        """

        address = invalid_prep.address
        index: int = self._index_of_prep(self._main_preps, address)
        if index < 0:
            return index

        invalid_prep_snapshot = self._main_preps[index]

        if len(self._sub_preps) == 0:
            self._main_preps.pop(index)
            Logger.warning(tag=self.TAG,
                           msg=f"Not enough sub P-Rep to replace an invalid main P-Rep")
        else:
            self._main_preps[index] = self._sub_preps.pop(0)
            Logger.info(
                tag=self.TAG,
                msg=f"Replace a main P-Rep: "
                    f"index={index} {address} -> {self._main_preps[index].address}")

        self._reduce_total_elected_prep_delegated(invalid_prep, invalid_prep_snapshot.delegated)
        del self._preps_dict[address]

        return index

    def _remove_invalid_sub_prep(self, invalid_prep: 'PRep') -> int:
        """Remove an invalid sub P-Rep from self._sub_preps

        :param invalid_prep: an invalid sub P-Rep
        :return: the index of an given address in self._sub_preps
            if not exists, returns -1
        """
        index: int = self._index_of_prep(self._sub_preps, invalid_prep.address)

        if index >= 0:
            invalid_prep_snapshot = self._sub_preps.pop(index)
            self._reduce_total_elected_prep_delegated(invalid_prep, invalid_prep_snapshot.delegated)
            del self._preps_dict[invalid_prep.address]

        return index

    def _reduce_total_elected_prep_delegated(self, invalid_prep: 'PRep', delegated: int):
        """Reduce total_elected_prep_delegated by the delegated amount of the given invalid P-Rep

        :param invalid_prep:
        :param delegated:
        :return:
        """

        # This code is preserved only for state backward compatibility.
        # After revision 7, B2 reward is not provided to the P-Rep
        # which got penalized for consecutive 660 blocks validation failure
        if invalid_prep.status != PRepStatus.ACTIVE \
                or invalid_prep.penalty != PenaltyReason.BLOCK_VALIDATION:
            self._total_elected_prep_delegated_snapshot -= delegated
            Logger.info(tag=self.TAG,
                        msg="total_elected_prep_delegated_snapshot is changed: "
                            f"delta={-delegated} "
                            f"total_elected_prep_delegated_snapshot={self._total_elected_prep_delegated_snapshot}")

        self._total_elected_prep_delegated -= delegated
        Logger.info(tag=self.TAG,
                    msg="total_elected_prep_delegated is changed: "
                        f"delta={-delegated} "
                        f"total_elected_prep_delegated={self._total_elected_prep_delegated}")

    @classmethod
    def _index_of_prep(cls, preps: List['PRepSnapshot'], address: 'Address') -> int:
        for i, snapshot in enumerate(preps):
            if address == snapshot.address:
                return i

        return -1

    def _generate_root_hash(self):
        def _gen(snapshots: Iterable['PRepSnapshot']) -> bytes:
            for snapshot in snapshots:
                yield snapshot.address.to_bytes_including_prefix()

        self._merkle_root_hash: bytes = \
            RootHashGenerator.generate_root_hash(values=_gen(self._main_preps), do_hash=True)

    @classmethod
    def from_list(cls, data: List,
                  block_height: int,
                  total_elected_prep_delegated_from_rc: int) -> 'Term':
        assert data[0] == cls._VERSION
        sequence: int = data[1]
        start_block_height: int = data[2]
        period: int = data[3]
        irep: int = data[4]
        total_supply: int = data[5]
        total_delegated: int = data[6]
        main_preps = data[7]
        sub_preps = data[8]
        total_elected_prep_delegated = 0

        term = Term(sequence, start_block_height, period, irep, total_supply, total_delegated)

        for address, delegated in main_preps:
            snapshot = PRepSnapshot(address, delegated)
            term._main_preps.append(snapshot)
            term._preps_dict[address] = snapshot
            total_elected_prep_delegated += delegated

        for address, delegated in sub_preps:
            snapshot = PRepSnapshot(address, delegated)
            term._sub_preps.append(snapshot)
            term._preps_dict[address] = snapshot
            total_elected_prep_delegated += delegated

        term._total_elected_prep_delegated = total_elected_prep_delegated

        if block_height == start_block_height - 1 or total_elected_prep_delegated_from_rc <= 0:
            # In the case of the first term (prevote -> decentralization),
            # total_elected_prep_delegated_from_rc can be 0.
            # and
            # fix IS-965 Sync fails on block height 10491442 when sync by using a master branch
            total_elected_prep_delegated_from_rc = total_elected_prep_delegated
        term._total_elected_prep_delegated_snapshot = total_elected_prep_delegated_from_rc

        term._generate_root_hash()

        return term

    def to_list(self) -> List:
        return [
            self._VERSION,
            self._sequence,
            self._start_block_height,
            self._period,
            self._irep,
            self._total_supply,
            self._total_delegated,
            [[snapshot.address, snapshot.delegated] for snapshot in self._main_preps],
            [[snapshot.address, snapshot.delegated] for snapshot in self._sub_preps],
        ]

    def copy(self) -> 'Term':
        term = copy.copy(self)
        term._is_frozen = False
        term._flags = TermFlag.NONE

        term._main_preps = list(self._main_preps)
        term._sub_preps = list(self._sub_preps)
        term._preps_dict = dict(self._preps_dict)

        return term
